import dataclasses
import hashlib
import math
from enum import Enum
from typing import Iterator, Tuple, TypeVar, List, Optional

import bitstruct

T = TypeVar("T")


def _bits_for_number(value: int) -> int:
    return int(math.ceil(math.log2(value)))


def single_byte_hash(data: bytes) -> int:
    return hashlib.blake2b(data, digest_size=1).digest()[0]


def _compile_format(*args):
    return bitstruct.CompiledFormat("".join("u{}".format(_bits_for_number(v)) for v in args))


class BitPackDecoder:
    _data: bytes
    _offset: int

    def __init__(self, data: bytes):
        self._data = data
        self._offset = 0

    def decode(self, *args) -> Tuple[int, ...]:
        """Decodes values from the current buffer, advancing the current pointer"""
        compiled = _compile_format(*args)
        offset = self._offset
        self._offset += compiled.calcsize()
        return compiled.unpack_from(self._data, offset)

    def decode_single(self, value: int) -> int:
        return self.decode(value)[0]

    def decode_element(self, array: List[T]) -> T:
        if len(array) == 1:
            return array[0]
        return array[self.decode_single(len(array))]

    def peek(self, *args) -> Tuple[int, ...]:
        """Decodes values from the current buffer, *NOT* advancing the current pointer"""
        compiled = _compile_format(*args)
        return compiled.unpack_from(self._data, self._offset)


class BitPackValue:
    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        raise NotImplementedError()

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata):
        raise NotImplementedError()

    @classmethod
    def bit_pack_skip_if_equals(cls):
        return True


class BitPackBool(BitPackValue):
    value: bool

    def __init__(self, value: bool):
        self.value = value

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        yield from encode_bool(self.value)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> bool:
        return decode_bool(decoder)

    @classmethod
    def bit_pack_skip_if_equals(cls):
        return False


class BitPackFloat(BitPackValue):
    value: float

    def __init__(self, value: float):
        self.value = value

    def bit_pack_encode(self, metadata: dict) -> Iterator[Tuple[int, int]]:
        if "if_different" in metadata:
            same = self.value == metadata["if_different"]
            yield from encode_bool(same)
            if same:
                return

        value_range = (metadata["max"] - metadata["min"]) * (10 ** metadata["precision"])
        yield int((self.value - metadata["min"]) * (10 ** metadata["precision"])), int(value_range) + 1

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> float:
        if "if_different" in metadata:
            same = decode_bool(decoder)
            if same:
                return metadata["if_different"]

        value_range = (metadata["max"] - metadata["min"]) * (10 ** metadata["precision"])
        decoded = decoder.decode_single(int(value_range) + 1)
        return float((decoded / (10 ** metadata["precision"])) + metadata["min"])


class BitPackInt(BitPackValue):
    value: int

    def __init__(self, value: int):
        self.value = value

    def bit_pack_encode(self, metadata: dict) -> Iterator[Tuple[int, int]]:
        if "if_different" in metadata:
            same = self.value == metadata["if_different"]
            yield from encode_bool(same)
            if same:
                return

        value_range = metadata["max"] - metadata["min"]
        yield self.value - metadata["min"], value_range

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> float:
        if "if_different" in metadata:
            same = decode_bool(decoder)
            if same:
                return metadata["if_different"]

        value_range = (metadata["max"] - metadata["min"])
        decoded = decoder.decode_single(value_range)
        return decoded + metadata["min"]


class BitPackEnum(BitPackValue):
    def __reduce__(self):
        return None

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        cls: Enum = self.__class__
        values = list(cls.__members__.values())
        yield from pack_array_element(self, values)

    @classmethod
    def bit_pack_unpack(cls: "Enum", decoder: BitPackDecoder, metadata):
        items = list(cls.__members__.values())
        return decoder.decode_element(items)


_default_bit_pack_classes = {
    bool: BitPackBool,
    int: BitPackInt,
    float: BitPackFloat,
}


def _get_bit_pack_value_for_type(value_type):
    if issubclass(value_type, BitPackValue):
        return value_type

    if value_type in _default_bit_pack_classes:
        return _default_bit_pack_classes[value_type]

    else:
        raise NotImplementedError("Unsupported bit packing for type {}".format(value_type))


def _get_bit_pack_value_for(value, dataclass_type: type):
    if isinstance(value, BitPackValue):
        return value

    return _get_bit_pack_value_for_type(dataclass_type)(value)


class BitPackDataClass(BitPackValue):
    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        reference = metadata.get("reference")

        for field in dataclasses.fields(self):
            if not field.init:
                continue

            item = getattr(self, field.name)
            bit_pack_value = _get_bit_pack_value_for(item, field.type)

            if reference is not None:
                reference_item = getattr(reference, field.name)
            else:
                reference_item = None

            field_meta = dict(**field.metadata)
            field_meta["reference"] = reference_item

            encoded_item = list(bit_pack_value.bit_pack_encode(field_meta))
            if any((a >= b) for (a, b) in encoded_item):
                raise ValueError(f"Encoding field {field.name} of {type(self)} generated invalid value.")
            should_encode = True

            if bit_pack_value.bit_pack_skip_if_equals() and reference_item is not None:
                reference_pack_value = _get_bit_pack_value_for(reference_item, field.type)
                encoded_reference = list(reference_pack_value.bit_pack_encode(field_meta))
                should_encode = encoded_item != encoded_reference
                yield from encode_bool(should_encode)

            if should_encode:
                yield from encoded_item

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata):
        reference = metadata.get("reference")
        args = {}

        for field in dataclasses.fields(cls):
            if not field.init:
                continue

            bit_pack_value = _get_bit_pack_value_for_type(field.type)

            should_decode = True
            if reference is not None:
                reference_item = getattr(reference, field.name)
                if bit_pack_value.bit_pack_skip_if_equals() and not decode_bool(decoder):
                    item = reference_item
                    should_decode = False
            else:
                reference_item = None

            if should_decode:
                field_meta = dict(**field.metadata)
                field_meta["reference"] = reference_item
                item = bit_pack_value.bit_pack_unpack(decoder, field_meta)

            args[field.name] = item

        return cls(**args)


def pack_array_element(element: T, array: List[T]) -> Iterator[Tuple[int, int]]:
    if len(array) > 1:
        yield array.index(element), len(array)
    else:
        if element not in array:
            raise ValueError(f"given element {element} is not in array of {len(array)}")


def _is_sorted(array: List[T]) -> bool:
    return array == list(sorted(array))


def _limits_for_size(remaining_size: int) -> Tuple[int, ...]:
    if remaining_size > 4:
        return 4, remaining_size
    elif remaining_size > 0:
        return remaining_size,
    else:
        return ()


def _aux_pack_sorted_array_elements(elements: List[T], array: List[T]) -> Iterator[Tuple[int, int]]:
    yield len(elements), len(array) + 1

    previous_index = 0
    for i, item in enumerate(elements):
        index = array.index(item)
        assert index is not None

        remaining_size = len(array) - previous_index - (len(elements) - i)
        yield from encode_int_with_limits(index - previous_index, _limits_for_size(remaining_size))
        previous_index = index


def pack_sorted_array_elements(elements: List[T], array: List[T]) -> Iterator[Tuple[int, int]]:
    assert _is_sorted(elements)
    assert _is_sorted(array)
    assert len(array) == len(set(array))

    if not array:
        return

    inverted_elements = [item for item in array if item not in elements]

    normal_results = list(_aux_pack_sorted_array_elements(elements, array))
    inverted_results = list(_aux_pack_sorted_array_elements(inverted_elements, array))

    bits_for_normal = sum(_bits_for_number(v[1]) for v in normal_results)
    bits_for_inverted = sum(_bits_for_number(v[1]) for v in inverted_results)

    should_invert = bits_for_normal > bits_for_inverted
    yield from encode_bool(should_invert)
    if should_invert:
        yield from inverted_results
    else:
        yield from normal_results


def decode_sorted_array_elements(decoder: BitPackDecoder, array: List[T]) -> List[T]:
    result = []

    if not array:
        return []

    inverted = decode_bool(decoder)
    elements_size = decoder.decode_single(len(array) + 1)

    previous_index = 0
    for i in range(elements_size):
        remaining_size = len(array) - previous_index - (elements_size - i)
        index = decode_int_with_limits(decoder, _limits_for_size(remaining_size))

        previous_index += index
        result.append(array[previous_index])

    if inverted:
        return [item for item in array if item not in result]
    else:
        return result


def encode_int_with_limits(value: int, limits: Tuple[int, ...]) -> Iterator[Tuple[int, int]]:
    previous_limit_sum = 0

    for i, limit in enumerate(limits):
        limit -= previous_limit_sum
        previous_limit_sum += limit

        if i == len(limits) - 1 or value < limit:
            yield value, limit + 1
            break

        yield limit, limit + 1
        value -= limit


def decode_int_with_limits(decoder: BitPackDecoder, limits: Tuple[int, ...]) -> int:
    value = 0
    previous_limit_sum = 0

    for limit in limits:
        limit -= previous_limit_sum
        previous_limit_sum += limit

        new_value = decoder.decode_single(limit + 1)
        value += new_value
        if new_value < limit:
            break

    return value


def encode_bytes(b: bytes):
    yield from encode_int_with_limits(len(b), (8, 32, 128))
    for item in b:
        yield item, 256


def decode_bytes(decoder: BitPackDecoder) -> bytes:
    size = decode_int_with_limits(decoder, (8, 32, 128))
    return bytes(decoder.decode(*([256] * size)))


def encode_string(s: str):
    yield from encode_bytes(s.encode("utf-8"))


def decode_string(decoder: BitPackDecoder) -> str:
    return decode_bytes(decoder).decode("utf-8")


def encode_bool(value: bool) -> Iterator[Tuple[int, int]]:
    yield int(value), 2


def decode_bool(decoder: BitPackDecoder) -> bool:
    return bool(decoder.decode_single(2))


def _format_string_for(values: List[Tuple[int, int]]) -> str:
    return "".join(
        "u{}".format(_bits_for_number(v))
        for _, v in values
    )


def _pack_encode_results(values: List[Tuple[int, int]]):
    f = _format_string_for(values)
    return bitstruct.compile(f).pack(*[argument for argument, _ in values])


def pack_value(value: BitPackValue, metadata: Optional[dict] = None) -> bytes:
    if metadata is None:
        metadata = {}

    results = []

    for i, (value_argument, value_format) in enumerate(value.bit_pack_encode(metadata)):
        if value_argument >= value_format:
            raise ValueError(f"At {i}, got {value_argument} which is bigger than limit {value_format}")
        results.append((value_argument, value_format))

    return _pack_encode_results(results)


def round_trip(value: T,
               metadata: Optional[dict] = None) -> T:
    """
    Encodes the given value and then recreates it using the encoded value
    :param value:
    :param metadata:
    :return:
    """
    if metadata is None:
        metadata = {}

    b = pack_value(value, metadata)
    decoder = BitPackDecoder(b)
    return value.__class__.bit_pack_unpack(decoder, metadata)



import json
import sys
from pathlib import Path
from typing import Union
from unittest.mock import patch, MagicMock, call, ANY

import pytest

import randovania
import randovania.generator.elevator_distributor
from randovania.game_description.game_patches import GamePatches
from randovania.games.prime import claris_randomizer
from randovania.interface_common.echoes_user_preferences import EchoesUserPreferences
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink

LayoutDescriptionMock = Union[MagicMock, LayoutDescription]


class CustomException(Exception):
    @classmethod
    def do_raise(cls, x):
        raise CustomException("test exception")


@pytest.fixture(
    params=[False, True],
    name="mock_is_windows")
def _mock_is_windows(request):
    with patch("randovania.games.prime.claris_randomizer._is_windows", return_value=request.param):
        yield request.param


@pytest.fixture(name="valid_tmp_game_root")
def _valid_tmp_game_root(tmp_path):
    game_root = tmp_path.joinpath("game_root")
    game_root.joinpath("files").mkdir(parents=True)
    game_root.joinpath("sys").mkdir()

    for f in ['default.dol', 'FrontEnd.pak', 'Metroid1.pak', 'Metroid2.pak']:
        game_root.joinpath("files", f).write_bytes(b"")

    game_root.joinpath("sys", 'main.dol').write_bytes(b"")

    return game_root


@patch("randovania.games.prime.claris_randomizer._process_command", autospec=True)
def test_run_with_args_success(mock_process_command: MagicMock,
                               mock_is_windows: bool,
                               ):
    # Setup
    args = [MagicMock(), MagicMock()]
    finish_string = "We are done!"
    status_update = MagicMock()
    lines = [
        "line 1",
        "line 2",
        finish_string,
        "post line"
    ]

    def side_effect(_, __, read_callback):
        for line in lines:
            read_callback(line)

    mock_process_command.side_effect = side_effect

    # Run
    claris_randomizer._run_with_args(args, "", finish_string, status_update)

    # Assert
    mock_process_command.assert_called_once_with(
        ([] if mock_is_windows else ["mono"]) + [str(x) for x in args], "", ANY)
    status_update.assert_has_calls([
        call("line 1"),
        call("line 2"),
        call(finish_string),
    ])


@patch("randovania.games.prime.claris_randomizer._process_command", autospec=True)
def test_run_with_args_failure(mock_process_command: MagicMock,
                               mock_is_windows: bool,
                               ):
    # Setup
    input_data = "asdf"
    finish_string = "We are done!"
    status_update = MagicMock()
    lines = [
        "line 1",
        "line 2",
        "post line"
    ]

    def side_effect(_, __, read_callback):
        for line in lines:
            read_callback(line)

    mock_process_command.side_effect = side_effect

    # Run
    with pytest.raises(RuntimeError) as error:
        claris_randomizer._run_with_args([], input_data, finish_string, status_update)

    # Assert
    mock_process_command.assert_called_once_with([] if mock_is_windows else ["mono"], input_data, ANY)
    status_update.assert_has_calls([
        call("line 1"),
        call("line 2"),
        call("post line"),
    ])
    assert str(error.value) == "External tool did not send '{}'. Did something happen?".format(finish_string)


@patch("randovania.games.prime.claris_randomizer.validate_game_files_path", autospec=True)
@patch("randovania.games.prime.claris_randomizer.get_data_path", autospec=True)
def test_base_args(mock_get_data_path: MagicMock,
                   mock_validate_game_files_path: MagicMock,
                   ):
    # Setup
    mock_get_data_path.return_value = Path("data")
    game_root = Path("root")

    # Run
    results = claris_randomizer._base_args(game_root)

    # Assert
    expected_results = [
        Path("data", "ClarisPrimeRandomizer", "Randomizer.exe"),
        Path("root"),
    ]

    assert results == expected_results
    mock_validate_game_files_path.assert_called_once_with(Path("root", "files"))


@pytest.mark.parametrize("has_menu_mod", [False, True])
@pytest.mark.parametrize("has_backup", [False, True, None])
@patch("shutil.copy", autospec=True)
def test_ensure_no_menu_mod(mock_copy: MagicMock,
                            tmpdir,
                            has_menu_mod: bool,
                            has_backup: bool,
                            ):
    # Setup
    game_root = Path(tmpdir.join("root"))
    backup_files_path = Path(tmpdir.join("backup"))
    status_update = MagicMock()
    files_folder = game_root.joinpath("files")
    mod_txt = files_folder.joinpath("menu_mod.txt")
    paks = ("1.pak", "2.pak")

    if has_menu_mod:
        mod_txt.parent.mkdir(parents=True)
        mod_txt.write_bytes(b"")

    if has_backup:
        pak_folder = backup_files_path.joinpath("mp2_paks")
        pak_folder.mkdir(parents=True)
        for pak in paks:
            pak_folder.joinpath(pak).write_bytes(b"")

    elif has_backup is None:
        backup_files_path = None

    # Run
    if has_menu_mod and has_backup is None:
        with pytest.raises(RuntimeError) as exc:
            claris_randomizer._ensure_no_menu_mod(game_root, backup_files_path, status_update)
        assert str(exc.value) == "Game at '{}' has Menu Mod, but no backup path given to restore".format(game_root)
    else:
        claris_randomizer._ensure_no_menu_mod(game_root, backup_files_path, status_update)

    # Assert
    if has_menu_mod:
        assert mod_txt.exists() != has_backup

    if has_menu_mod and has_backup:
        # In any order, because file systems don't have guaranteed order
        status_update.assert_has_calls([
            call("Restoring {} from backup".format(pak_name))
            for pak_name in paks
        ], any_order=True)
        mock_copy.assert_has_calls([
            call(backup_files_path.joinpath("mp2_paks", pak_name), files_folder.joinpath(pak_name))
            for pak_name in paks
        ], any_order=True)
    else:
        status_update.assert_not_called()
        mock_copy.assert_not_called()


@pytest.mark.parametrize("missing_pak", claris_randomizer._ECHOES_PAKS)
@patch("shutil.copy", autospec=True)
def test_create_pak_backups(mock_copy: MagicMock,
                            tmpdir,
                            missing_pak: str
                            ):
    # Setup
    game_root = Path(tmpdir.join("root"))
    backup_files_path = Path(tmpdir.join("backup"))
    status_update = MagicMock()

    pak_folder = backup_files_path.joinpath("mp2_paks")
    pak_folder.mkdir(parents=True)
    for pak in claris_randomizer._ECHOES_PAKS:
        if pak != missing_pak:
            pak_folder.joinpath(pak).write_bytes(b"")

    # Run
    claris_randomizer._create_pak_backups(game_root, backup_files_path, status_update)

    # Assert
    status_update.assert_called_once_with("Backing up {}".format(missing_pak))
    mock_copy.assert_called_once_with(game_root.joinpath("files", missing_pak),
                                      pak_folder.joinpath(missing_pak))


@patch("shutil.copy", autospec=True)
def test_create_pak_backups_no_existing(mock_copy: MagicMock,
                                        tmpdir,
                                        ):
    # Setup
    game_root = Path(tmpdir.join("root"))
    backup_files_path = Path(tmpdir.join("backup"))
    status_update = MagicMock()

    # Run
    claris_randomizer._create_pak_backups(game_root, backup_files_path, status_update)

    # Assert
    status_update.assert_has_calls([
        call("Backing up {}".format(pak))
        for pak in claris_randomizer._ECHOES_PAKS
    ])
    mock_copy.assert_has_calls([
        call(game_root.joinpath("files", pak), backup_files_path.joinpath("mp2_paks", pak))
        for pak in claris_randomizer._ECHOES_PAKS
    ])


@patch("randovania.games.prime.claris_randomizer._run_with_args", autospec=True)
@patch("randovania.games.prime.claris_randomizer.get_data_path", autospec=True)
def test_add_menu_mod_to_files(mock_get_data_path: MagicMock,
                               mock_run_with_args: MagicMock,
                               tmpdir,
                               ):
    # Setup
    mock_get_data_path.return_value = Path("data")
    game_root = Path(tmpdir.join("root"))
    status_update = MagicMock()
    game_root.joinpath("files").mkdir(parents=True)

    # Run
    claris_randomizer._add_menu_mod_to_files(game_root, status_update)

    # Assert
    mock_run_with_args.assert_called_once_with(
        [Path("data", "ClarisEchoesMenu", "EchoesMenu.exe"), game_root.joinpath("files")],
        "",
        "Done!",
        status_update
    )
    assert game_root.joinpath("files", "menu_mod.txt").is_file()


@pytest.mark.parametrize("include_menu_mod", [False, True])
@pytest.mark.parametrize("has_backup_path", [False, True])
@patch("randovania.interface_common.status_update_lib.create_progress_update_from_successive_messages", autospec=True)
@patch("randovania.games.prime.dol_patcher.apply_patches", autospec=True)
@patch("randovania.games.prime.claris_randomizer._run_with_args", autospec=True)
@patch("randovania.games.prime.claris_randomizer._add_menu_mod_to_files", autospec=True)
@patch("randovania.games.prime.claris_randomizer._create_pak_backups", autospec=True)
@patch("randovania.games.prime.claris_randomizer._ensure_no_menu_mod", autospec=True)
def test_apply_patcher_file(
        mock_ensure_no_menu_mod: MagicMock,
        mock_create_pak_backups: MagicMock,
        mock_add_menu_mod_to_files: MagicMock,
        mock_run_with_args: MagicMock,
        mock_apply_patches: MagicMock,
        mock_create_progress_update_from_successive_messages: MagicMock,
        include_menu_mod: bool,
        has_backup_path: bool,
        valid_tmp_game_root
):
    # Setup
    game_root = valid_tmp_game_root
    backup_files_path = MagicMock() if has_backup_path else None
    game_specific = MagicMock()
    progress_update = MagicMock()
    status_update = mock_create_progress_update_from_successive_messages.return_value

    patcher_data = {
        "menu_mod": include_menu_mod,
        "user_preferences": EchoesUserPreferences().as_json,
        "default_items": {"foo": "bar"},
    }
    assert claris_randomizer.get_patch_version(game_root) == 0

    # Run
    claris_randomizer.apply_patcher_file(game_root, backup_files_path,
                                         patcher_data, game_specific, progress_update)

    # Assert
    mock_create_progress_update_from_successive_messages.assert_called_once_with(
        progress_update,
        400 if include_menu_mod else 100
    )
    mock_ensure_no_menu_mod.assert_called_once_with(game_root, backup_files_path, status_update)
    if has_backup_path:
        mock_create_pak_backups.assert_called_once_with(game_root, backup_files_path, status_update)
    else:
        mock_create_pak_backups.assert_not_called()
    mock_run_with_args.assert_called_once_with(claris_randomizer._base_args(game_root),
                                               json.dumps(patcher_data), "Randomized!", status_update)
    mock_apply_patches.assert_called_once_with(game_root, game_specific, EchoesUserPreferences(), {"foo": "bar"})

    if include_menu_mod:
        mock_add_menu_mod_to_files.assert_called_once_with(game_root, status_update)
    else:
        mock_add_menu_mod_to_files.assert_not_called()

    assert claris_randomizer.get_patch_version(game_root) == claris_randomizer.CURRENT_PATCH_VERSION


@patch("randovania.games.prime.patcher_file.create_patcher_file", autospec=True)
@patch("randovania.games.prime.claris_randomizer.apply_patcher_file", autospec=True)
def test_apply_layout(mock_apply_patcher_file: MagicMock,
                      mock_create_patcher_file: MagicMock,
                      ):
    # Setup
    description = MagicMock()
    description.file_extension.return_value = "foobar"
    player_config = MagicMock()
    cosmetic_patches = MagicMock()
    backup_files_path = MagicMock()
    progress_update = MagicMock()
    game_root = MagicMock(spec=Path())

    # Run
    claris_randomizer.apply_layout(
        description, player_config, cosmetic_patches,
        backup_files_path, progress_update, game_root
    )

    # Assert
    game_root.joinpath.assert_called_once_with("files", "randovania.foobar")
    description.save_to_file.assert_called_once_with(game_root.joinpath.return_value)
    mock_create_patcher_file.assert_called_once_with(description, player_config, cosmetic_patches)
    mock_apply_patcher_file.assert_called_once_with(game_root, backup_files_path,
                                                    mock_create_patcher_file.return_value,
                                                    description.all_patches[player_config.player_index].game_specific,
                                                    progress_update)


def test_process_command_no_thread(echo_tool):
    read_callback = MagicMock()

    claris_randomizer.IO_LOOP = None

    # Run
    claris_randomizer._process_command(
        [
            sys.executable,
            str(echo_tool)
        ],
        "hello\r\nthis is a nice world\r\n\r\nWe some crazy stuff.",
        read_callback
    )

    # Assert
    read_callback.assert_has_calls([
        call("hello"),
        call("this is a nice world"),
        call("We some crazy stuff."),
    ])



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_beam_configuration.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_PresetEchoesBeamConfiguration(object):
    def setupUi(self, PresetEchoesBeamConfiguration):
        if not PresetEchoesBeamConfiguration.objectName():
            PresetEchoesBeamConfiguration.setObjectName(u"PresetEchoesBeamConfiguration")
        PresetEchoesBeamConfiguration.resize(466, 454)
        self.centralWidget = QWidget(PresetEchoesBeamConfiguration)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.beam_configuration_tab_layout = QVBoxLayout(self.centralWidget)
        self.beam_configuration_tab_layout.setSpacing(6)
        self.beam_configuration_tab_layout.setContentsMargins(11, 11, 11, 11)
        self.beam_configuration_tab_layout.setObjectName(u"beam_configuration_tab_layout")
        self.beam_configuration_tab_layout.setContentsMargins(4, 4, 4, 0)
        self.beam_configuration_label = QLabel(self.centralWidget)
        self.beam_configuration_label.setObjectName(u"beam_configuration_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.beam_configuration_label.sizePolicy().hasHeightForWidth())
        self.beam_configuration_label.setSizePolicy(sizePolicy)
        self.beam_configuration_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.beam_configuration_label.setWordWrap(True)

        self.beam_configuration_tab_layout.addWidget(self.beam_configuration_label)

        self.beam_configuration_group = QGroupBox(self.centralWidget)
        self.beam_configuration_group.setObjectName(u"beam_configuration_group")
        self.beam_configuration_group.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.beam_configuration_group.setFlat(True)
        self.beam_configuration_layout = QGridLayout(self.beam_configuration_group)
        self.beam_configuration_layout.setSpacing(6)
        self.beam_configuration_layout.setContentsMargins(11, 11, 11, 11)
        self.beam_configuration_layout.setObjectName(u"beam_configuration_layout")
        self.beam_configuration_layout.setContentsMargins(0, 0, 0, 0)

        self.beam_configuration_tab_layout.addWidget(self.beam_configuration_group)

        self.beam_configuration_spacer = QSpacerItem(20, 311, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.beam_configuration_tab_layout.addItem(self.beam_configuration_spacer)

        PresetEchoesBeamConfiguration.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesBeamConfiguration)

        QMetaObject.connectSlotsByName(PresetEchoesBeamConfiguration)
    # setupUi

    def retranslateUi(self, PresetEchoesBeamConfiguration):
        PresetEchoesBeamConfiguration.setWindowTitle(QCoreApplication.translate("PresetEchoesBeamConfiguration", u"Beam Configuration", None))
        self.beam_configuration_label.setText(QCoreApplication.translate("PresetEchoesBeamConfiguration", u"<html><head/><body><p>Configure what each beam uses as ammo and how much ammo is consumed when shooting it uncharged, charged or with a charge combo.</p><p>Logic always uses the default values.</p><p><span style=\" font-weight:600;\">Known Issue: </span>If you're out of the ammo a beam normally uses, you'll be unable to shoot that beam regardless of what it actually uses to shoot.</p></body></html>", None))
        self.beam_configuration_group.setTitle("")
    # retranslateUi




from enum import Enum

from randovania.bitpacking.bitpacking import BitPackEnum


class LayoutDamageStrictness(BitPackEnum, Enum):
    STRICT = 1.0
    MEDIUM = 1.5
    LENIENT = 2.0

    @classmethod
    def default(cls):
        return cls.MEDIUM

    @property
    def long_name(self) -> str:
        if self == LayoutDamageStrictness.STRICT:
            return "Strict"
        elif self == LayoutDamageStrictness.MEDIUM:
            return "Medium"
        elif self == LayoutDamageStrictness.LENIENT:
            return "Lenient"
        else:
            return "Custom"



import cryptography.fernet
import flask
import pytest
from peewee import SqliteDatabase

from randovania.server import database


@pytest.fixture()
def clean_database():
    old_db = database.db
    try:
        test_db = SqliteDatabase(':memory:')
        database.db = test_db
        with test_db.bind_ctx(database.all_classes):
            test_db.connect(reuse_if_open=True)
            test_db.create_tables(database.all_classes)
            yield test_db
    finally:
        database.db = old_db


@pytest.fixture()
def flask_app():
    app = flask.Flask("test_app")
    app.config["TESTING"] = True
    app.config["SECRET_KEY"] = "secret"
    with app.app_context():
        yield app


@pytest.fixture()
def fernet():
    return cryptography.fernet.Fernet(b's2D-pjBIXqEqkbeRvkapeDn82MgZXLLQGZLTgqqZ--A=')



import collections
import functools
import json
import typing
from pathlib import Path
from random import Random
from typing import Optional, Dict, Set, List, Tuple, Iterator, Union

import matplotlib.pyplot as plt
import networkx
from PySide2 import QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QMainWindow, QTreeWidgetItem, QCheckBox, QLabel, QGridLayout, QWidget, QMessageBox
from matplotlib.axes import Axes
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure

from randovania.game_description.area_location import AreaLocation
from randovania.game_description.game_description import GameDescription
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.node import Node, ResourceNode, TranslatorGateNode, TeleporterNode, DockNode
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_info import add_resource_gain_to_current_resources
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.world import World
from randovania.games.game import RandovaniaGame
from randovania.generator import generator
from randovania.gui.generated.tracker_window_ui import Ui_TrackerWindow
from randovania.gui.lib.common_qt_lib import set_default_window_icon
from randovania.gui.lib.custom_spin_box import CustomSpinBox
from randovania.layout import translator_configuration
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.elevators import LayoutElevators
from randovania.layout.translator_configuration import LayoutTranslatorRequirement
from randovania.resolver.bootstrap import logic_bootstrap
from randovania.resolver.logic import Logic
from randovania.resolver.resolver_reach import ResolverReach
from randovania.resolver.state import State, add_pickup_to_state


class InvalidLayoutForTracker(Exception):
    pass


def _load_previous_state(persistence_path: Path,
                         layout_configuration: EchoesConfiguration,
                         ) -> Optional[dict]:
    previous_layout_path = persistence_path.joinpath("layout_configuration.json")
    try:
        with previous_layout_path.open() as previous_layout_file:
            previous_layout = EchoesConfiguration.from_json(json.load(previous_layout_file))
    except (FileNotFoundError, TypeError, KeyError, ValueError, json.JSONDecodeError):
        return None

    if previous_layout != layout_configuration:
        return None

    previous_state_path = persistence_path.joinpath("state.json")
    try:
        with previous_state_path.open() as previous_state_file:
            return json.load(previous_state_file)
    except (FileNotFoundError, json.JSONDecodeError):
        return None


class MatplotlibWidget(QtWidgets.QWidget):
    ax: Axes

    def __init__(self, parent=None):
        super().__init__(parent)

        fig = Figure(figsize=(7, 5), dpi=65, facecolor=(1, 1, 1), edgecolor=(0, 0, 0))
        self.canvas = FigureCanvas(fig)
        self.toolbar = NavigationToolbar(self.canvas, self)
        lay = QtWidgets.QVBoxLayout(self)
        lay.addWidget(self.toolbar)
        lay.addWidget(self.canvas)

        self.ax = fig.add_subplot(111)
        self.line, *_ = self.ax.plot([])


class TrackerWindow(QMainWindow, Ui_TrackerWindow):
    # Tracker state
    _collected_pickups: Dict[PickupEntry, int]
    _actions: List[Node]

    # Tracker configuration
    logic: Logic
    game_description: GameDescription
    layout_configuration: EchoesConfiguration
    persistence_path: Path
    _initial_state: State
    _elevator_id_to_combo: Dict[int, QtWidgets.QComboBox]
    _translator_gate_to_combo: Dict[TranslatorGate, QtWidgets.QComboBox]
    _starting_nodes: Set[ResourceNode]
    _undefined_item = ItemResourceInfo(-1, "Undefined", "Undefined", 0, None)

    # UI tools
    _asset_id_to_item: Dict[int, QTreeWidgetItem]
    _node_to_item: Dict[Node, QTreeWidgetItem]
    _widget_for_pickup: Dict[PickupEntry, Union[QCheckBox, CustomSpinBox]]
    _during_setup = False

    def __init__(self, persistence_path: Path, layout_configuration: EchoesConfiguration):
        super().__init__()
        self.setupUi(self)
        set_default_window_icon(self)

        self._collected_pickups = {}
        self._widget_for_pickup = {}
        self._actions = []
        self._asset_id_to_item = {}
        self._node_to_item = {}
        self.layout_configuration = layout_configuration
        self.persistence_path = persistence_path

        player_pool = generator.create_player_pool(Random(0), self.layout_configuration, 0, 1)
        pool_patches = player_pool.patches
        self.game_description, self._initial_state = logic_bootstrap(layout_configuration,
                                                                     player_pool.game,
                                                                     pool_patches)
        self.logic = Logic(self.game_description, layout_configuration)

        self._initial_state.resources["add_self_as_requirement_to_resources"] = 1

        self.menu_reset_action.triggered.connect(self._confirm_reset)
        self.resource_filter_check.stateChanged.connect(self.update_locations_tree_for_reachable_nodes)
        self.hide_collected_resources_check.stateChanged.connect(self.update_locations_tree_for_reachable_nodes)
        self.undo_last_action_button.clicked.connect(self._undo_last_action)

        self.configuration_label.setText("Trick Level: {}; Starts with:\n{}".format(
            layout_configuration.trick_level.pretty_description,
            ", ".join(
                resource.short_name
                for resource in pool_patches.starting_items.keys()
            )
        ))

        self.setup_pickups_box(player_pool.pickups)
        self.setup_possible_locations_tree()
        self.setup_elevators()
        self.setup_translator_gates()

        self.matplot_widget = MatplotlibWidget(self.tab_graph_map)
        self.tab_graph_map_layout.addWidget(self.matplot_widget)
        self._world_to_node_positions = {}
        self.map_tab_widget.currentChanged.connect(self._on_tab_changed)

        for world in self.game_description.world_list.worlds:
            self.graph_map_world_combo.addItem(world.name, world)
        self.graph_map_world_combo.currentIndexChanged.connect(self.on_graph_map_world_combo)

        persistence_path.mkdir(parents=True, exist_ok=True)
        previous_state = _load_previous_state(persistence_path, layout_configuration)

        if not self.apply_previous_state(previous_state):
            self.setup_starting_location(None)

            with persistence_path.joinpath("layout_configuration.json").open("w") as layout_file:
                json.dump(layout_configuration.as_json, layout_file)
            self._add_new_action(self._initial_state.node)

    def apply_previous_state(self, previous_state: Optional[dict]) -> bool:
        if previous_state is None:
            return False

        starting_location = None
        needs_starting_location = len(self.layout_configuration.starting_location.locations) > 1
        resource_db = self.game_description.resource_database
        translator_gates = {}

        try:
            pickup_name_to_pickup = {pickup.name: pickup for pickup in self._collected_pickups.keys()}
            quantity_to_change = {
                pickup_name_to_pickup[pickup_name]: quantity
                for pickup_name, quantity in previous_state["collected_pickups"].items()
            }
            previous_actions = [
                self.game_description.world_list.all_nodes[index]
                for index in previous_state["actions"]
            ]
            if needs_starting_location:
                starting_location = AreaLocation.from_json(previous_state["starting_location"])

            elevators = {
                int(elevator_id): AreaLocation.from_json(location) if location is not None else None
                for elevator_id, location in previous_state["elevators"].items()
            }
            if self.layout_configuration.game == RandovaniaGame.PRIME2:
                translator_gates = {
                    TranslatorGate(int(gate)): (resource_db.get_item(item)
                                                if item is not None
                                                else self._undefined_item)
                    for gate, item in previous_state["translator_gates"].items()
                }
        except KeyError:
            return False

        self.setup_starting_location(starting_location)

        for elevator_id, area_location in elevators.items():
            combo = self._elevator_id_to_combo[elevator_id]
            if area_location is None:
                combo.setCurrentIndex(0)
                continue
            for i in range(combo.count()):
                if area_location == combo.itemData(i):
                    combo.setCurrentIndex(i)
                    break

        for gate, item in translator_gates.items():
            combo = self._translator_gate_to_combo[gate]
            for i in range(combo.count()):
                if item == combo.itemData(i):
                    combo.setCurrentIndex(i)
                    break

        self.bulk_change_quantity(quantity_to_change)
        self._add_new_actions(previous_actions)
        return True

    def reset(self):
        self.bulk_change_quantity({
            pickup: 0
            for pickup in self._collected_pickups.keys()
        })

        while len(self._actions) > 1:
            self._actions.pop()
            self.actions_list.takeItem(len(self._actions))

        for elevator in self._elevator_id_to_combo.values():
            elevator.setCurrentIndex(0)
        for elevator in self._translator_gate_to_combo.values():
            elevator.setCurrentIndex(0)

        self._refresh_for_new_action()

    def _confirm_reset(self):
        reply = QMessageBox.question(self, "Reset Tracker?", "Do you want to reset the tracker progression?",
                                     QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.reset()

    @property
    def _show_only_resource_nodes(self) -> bool:
        return self.resource_filter_check.isChecked()

    @property
    def _hide_collected_resources(self) -> bool:
        return self.hide_collected_resources_check.isChecked()

    @property
    def _collected_nodes(self) -> Set[ResourceNode]:
        return self._starting_nodes | set(action for action in self._actions if action.is_resource_node)

    def _pretty_node_name(self, node: Node) -> str:
        world_list = self.game_description.world_list
        return "{} / {}".format(world_list.area_name(world_list.nodes_to_area(node)), node.name)

    def _refresh_for_new_action(self):
        self.undo_last_action_button.setEnabled(len(self._actions) > 1)
        self.current_location_label.setText("Current location: {}".format(self._pretty_node_name(self._actions[-1])))
        self.update_locations_tree_for_reachable_nodes()

    def _add_new_action(self, node: Node):
        self._add_new_actions([node])

    def _add_new_actions(self, nodes: Iterator[Node]):
        for node in nodes:
            self.actions_list.addItem(self._pretty_node_name(node))
            self._actions.append(node)
        self._refresh_for_new_action()

    def _undo_last_action(self):
        self._actions.pop()
        self.actions_list.takeItem(len(self._actions))
        self._refresh_for_new_action()

    def _on_tree_node_double_clicked(self, item: QTreeWidgetItem, _):
        node: Optional[Node] = getattr(item, "node", None)

        if not item.isDisabled() and node is not None and node != self._actions[-1]:
            self._add_new_action(node)

    def _positions_for_world(self, world: World):
        g = networkx.DiGraph()
        world_list = self.game_description.world_list
        state = self.state_for_current_configuration()

        for area in world.areas:
            g.add_node(area)

        for area in world.areas:
            nearby_areas = set()
            for node in area.nodes:
                if isinstance(node, DockNode):
                    try:
                        target_node = world_list.resolve_dock_node(node, state.patches)
                        nearby_areas.add(world_list.nodes_to_area(target_node))
                    except IndexError as e:
                        print(f"For {node.name} in {area.name}, received {e}")
                        continue
            for other_area in nearby_areas:
                g.add_edge(area, other_area)

        return networkx.drawing.spring_layout(g)

    def update_matplot_widget(self, nodes_in_reach: Set[Node]):
        g = networkx.DiGraph()
        world_list = self.game_description.world_list
        state = self.state_for_current_configuration()

        world = self.graph_map_world_combo.currentData()
        for area in world.areas:
            g.add_node(area)

        for area in world.areas:
            nearby_areas = set()
            for node in area.nodes:
                if node not in nodes_in_reach:
                    continue

                if isinstance(node, DockNode):
                    # TODO: respect is_blast_shield: if already opened once, no requirement needed.
                    # Includes opening form behind with different criteria
                    try:
                        target_node = world_list.resolve_dock_node(node, state.patches)
                        dock_weakness = state.patches.dock_weakness.get((area.area_asset_id, node.dock_index),
                                                                        node.default_dock_weakness)
                        if dock_weakness.requirement.satisfied(state.resources, state.energy):
                            nearby_areas.add(world_list.nodes_to_area(target_node))
                    except IndexError as e:
                        print(f"For {node.name} in {area.name}, received {e}")
                        continue

            for other_area in nearby_areas:
                g.add_edge(area, other_area)

        self.matplot_widget.ax.clear()

        cf = self.matplot_widget.ax.get_figure()
        cf.set_facecolor("w")

        if world.world_asset_id not in self._world_to_node_positions:
            self._world_to_node_positions[world.world_asset_id] = self._positions_for_world(world)
        pos = self._world_to_node_positions[world.world_asset_id]

        networkx.draw_networkx_nodes(g, pos, ax=self.matplot_widget.ax)
        networkx.draw_networkx_edges(g, pos, arrows=True, ax=self.matplot_widget.ax)
        networkx.draw_networkx_labels(g, pos, ax=self.matplot_widget.ax,
                                      labels={area: area.name for area in world.areas},
                                      verticalalignment='top')

        self.matplot_widget.ax.set_axis_off()

        plt.draw_if_interactive()
        self.matplot_widget.canvas.draw()

    def on_graph_map_world_combo(self):
        nodes_in_reach = self.current_nodes_in_reach(self.state_for_current_configuration())
        self.update_matplot_widget(nodes_in_reach)

    def current_nodes_in_reach(self, state):
        if state is None:
            nodes_in_reach = set()
        else:
            reach = ResolverReach.calculate_reach(self.logic, state)
            nodes_in_reach = set(reach.nodes)
            nodes_in_reach.add(state.node)
        return nodes_in_reach

    def _on_tab_changed(self):
        if self.map_tab_widget.currentWidget() == self.tab_graph_map:
            self.on_graph_map_world_combo()

    def update_locations_tree_for_reachable_nodes(self):
        state = self.state_for_current_configuration()
        nodes_in_reach = self.current_nodes_in_reach(state)
        if self.map_tab_widget.currentWidget() == self.tab_graph_map:
            self.update_matplot_widget(nodes_in_reach)

        all_nodes = self.game_description.world_list.all_nodes
        for world in self.game_description.world_list.worlds:
            for area in world.areas:
                area_is_visible = False
                for node in area.nodes:
                    is_collected = node in self._collected_nodes
                    is_visible = node in nodes_in_reach and not (self._hide_collected_resources
                                                                 and is_collected)

                    if self._show_only_resource_nodes:
                        is_visible = is_visible and node.is_resource_node

                    node_item = self._node_to_item[node]
                    node_item.setHidden(not is_visible)
                    if node.is_resource_node:
                        resource_node = typing.cast(ResourceNode, node)
                        node_item.setDisabled(not resource_node.can_collect(state.patches, state.resources, all_nodes))
                        node_item.setCheckState(0, Qt.Checked if is_collected else Qt.Unchecked)

                    area_is_visible = area_is_visible or is_visible
                self._asset_id_to_item[area.area_asset_id].setHidden(not area_is_visible)

        # Persist the current state
        self.persist_current_state()

    def persist_current_state(self):
        world_list = self.game_description.world_list
        with self.persistence_path.joinpath("state.json").open("w") as state_file:
            json.dump(
                {
                    "actions": [
                        node.index
                        for node in self._actions
                    ],
                    "collected_pickups": {
                        pickup.name: quantity
                        for pickup, quantity in self._collected_pickups.items()
                    },
                    "elevators": {
                        str(elevator_id): combo.currentData().as_json if combo.currentIndex() > 0 else None
                        for elevator_id, combo in self._elevator_id_to_combo.items()
                    },
                    "translator_gates": {
                        str(gate.index): combo.currentData().index if combo.currentIndex() > 0 else None
                        for gate, combo in self._translator_gate_to_combo.items()
                    },
                    "starting_location": world_list.node_to_area_location(self._initial_state.node).as_json,
                },
                state_file
            )

    def setup_possible_locations_tree(self):
        """
        Creates the possible_locations_tree with all worlds, areas and nodes.
        """
        self.possible_locations_tree.itemDoubleClicked.connect(self._on_tree_node_double_clicked)

        # TODO: Dark World names
        for world in self.game_description.world_list.worlds:
            world_item = QTreeWidgetItem(self.possible_locations_tree)
            world_item.setText(0, world.name)
            world_item.setExpanded(True)
            self._asset_id_to_item[world.world_asset_id] = world_item

            for area in world.areas:
                area_item = QTreeWidgetItem(world_item)
                area_item.area = area
                area_item.setText(0, area.name)
                area_item.setHidden(True)
                self._asset_id_to_item[area.area_asset_id] = area_item

                for node in area.nodes:
                    node_item = QTreeWidgetItem(area_item)
                    if isinstance(node, TranslatorGateNode):
                        node_item.setText(0, "{} ({})".format(node.name, node.gate))
                    else:
                        node_item.setText(0, node.name)
                    node_item.node = node
                    if node.is_resource_node:
                        node_item.setFlags(node_item.flags() & ~Qt.ItemIsUserCheckable)
                    self._node_to_item[node] = node_item

    def setup_elevators(self):
        world_list = self.game_description.world_list
        nodes_by_world: Dict[str, List[TeleporterNode]] = collections.defaultdict(list)
        self._elevator_id_to_combo = {}

        areas_to_not_change = {
            2278776548,  # Sky Temple Gateway
            2068511343,  # Sky Temple Energy Controller
            3136899603,  # Aerie Transport Station
            1564082177,  # Aerie
        }
        targets = {}
        for world, area, node in world_list.all_worlds_areas_nodes:
            if isinstance(node, TeleporterNode) and node.editable and area.area_asset_id not in areas_to_not_change:
                name = world.correct_name(area.in_dark_aether)
                nodes_by_world[name].append(node)
                targets[f"{name} - {area.name}"] = AreaLocation(world.world_asset_id, area.area_asset_id)

        if self.layout_configuration.elevators == LayoutElevators.ONE_WAY_ANYTHING:
            targets = {}
            for world in world_list.worlds:
                for area in world.areas:
                    name = world.correct_name(area.in_dark_aether)
                    targets[f"{name} - {area.name}"] = AreaLocation(world.world_asset_id, area.area_asset_id)

        combo_targets = sorted(targets.items(), key=lambda it: it[0])

        for world_name in sorted(nodes_by_world.keys()):
            nodes = sorted(nodes_by_world[world_name], key=lambda it: world_list.nodes_to_area(it).name)

            group = QtWidgets.QGroupBox(self.elevators_scroll_contents)
            group.setTitle(world_name)
            self.elevators_scroll_layout.addWidget(group)
            layout = QtWidgets.QGridLayout(group)

            for i, node in enumerate(nodes):
                area = world_list.nodes_to_area(node)
                node_name = QtWidgets.QLabel(group)
                node_name.setText(area.name)
                layout.addWidget(node_name, i, 0)

                combo = QtWidgets.QComboBox(group)
                if self.layout_configuration.elevators == LayoutElevators.VANILLA:
                    combo.addItem("Vanilla", node.default_connection)
                    combo.setEnabled(False)
                else:
                    combo.addItem("Undefined", AreaLocation(world_list.nodes_to_world(node).world_asset_id,
                                                            area.area_asset_id))
                    for name, connection in combo_targets:
                        combo.addItem(name, connection)

                combo.currentIndexChanged.connect(self.update_locations_tree_for_reachable_nodes)
                self._elevator_id_to_combo[node.teleporter_instance_id] = combo
                layout.addWidget(combo, i, 1)

    def setup_translator_gates(self):
        world_list = self.game_description.world_list
        resource_db = self.game_description.resource_database
        self._translator_gate_to_combo = {}

        if self.layout_configuration.game != RandovaniaGame.PRIME2:
            return

        gates = {
            f"{area.name} ({node.gate.index})": node.gate
            for world, area, node in world_list.all_worlds_areas_nodes
            if isinstance(node, TranslatorGateNode)
        }
        translator_requirement = self.layout_configuration.translator_configuration.translator_requirement

        for i, (gate_name, gate) in enumerate(sorted(gates.items(), key=lambda it: it[0])):
            node_name = QtWidgets.QLabel(self.translator_gate_scroll_contents)
            node_name.setText(gate_name)
            self.translator_gate_scroll_layout.addWidget(node_name, i, 0)

            combo = QtWidgets.QComboBox(self.translator_gate_scroll_contents)
            gate_requirement = translator_requirement[gate]

            if gate_requirement in (LayoutTranslatorRequirement.RANDOM,
                                    LayoutTranslatorRequirement.RANDOM_WITH_REMOVED):
                combo.addItem("Undefined", self._undefined_item)
                for translator, index in translator_configuration.ITEM_INDICES.items():
                    combo.addItem(translator.long_name, resource_db.get_item(index))
            else:
                combo.addItem(gate_requirement.long_name, resource_db.get_item(gate_requirement.item_index))
                combo.setEnabled(False)

            combo.currentIndexChanged.connect(self.update_locations_tree_for_reachable_nodes)
            self._translator_gate_to_combo[gate] = combo
            self.translator_gate_scroll_layout.addWidget(combo, i, 1)

    def setup_starting_location(self, area_location: Optional[AreaLocation]):
        world_list = self.game_description.world_list

        if len(self.layout_configuration.starting_location.locations) > 1:
            if area_location is None:
                area_locations = sorted(self.layout_configuration.starting_location.locations,
                                        key=lambda it: world_list.area_name(world_list.area_by_area_location(it)))

                location_names = [world_list.area_name(world_list.area_by_area_location(it))
                                  for it in area_locations]
                selected_name = QtWidgets.QInputDialog.getItem(self, "Starting Location", "Select starting location",
                                                               location_names, 0, False)
                area_location = area_locations[location_names.index(selected_name[0])]

            self._initial_state.node = world_list.resolve_teleporter_connection(area_location)

        self._starting_nodes = {
            node
            for node in world_list.all_nodes
            if node.is_resource_node and node.resource() in self._initial_state.resources
        }

    def _change_item_quantity(self, pickup: PickupEntry, use_quantity_as_bool: bool, quantity: int):
        if use_quantity_as_bool:
            if bool(quantity):
                quantity = 1
            else:
                quantity = 0

        self._collected_pickups[pickup] = quantity
        if not self._during_setup:
            self.update_locations_tree_for_reachable_nodes()

    def bulk_change_quantity(self, new_quantity: Dict[PickupEntry, int]):
        self._during_setup = True
        for pickup, quantity in new_quantity.items():
            widget = self._widget_for_pickup[pickup]
            if isinstance(widget, QCheckBox):
                widget.setChecked(quantity > 0)
            else:
                widget.setValue(quantity)
        self._during_setup = False

    def _create_widgets_with_quantity(self,
                                      pickup: PickupEntry,
                                      parent_widget: QWidget,
                                      parent_layout: QGridLayout,
                                      row: int,
                                      quantity: int,
                                      ):
        label = QLabel(parent_widget)
        label.setText(pickup.name)
        parent_layout.addWidget(label, row, 0)

        spin_bix = CustomSpinBox(parent_widget)
        spin_bix.setMaximumWidth(50)
        spin_bix.setMaximum(quantity)
        spin_bix.valueChanged.connect(functools.partial(self._change_item_quantity, pickup, False))
        self._widget_for_pickup[pickup] = spin_bix
        parent_layout.addWidget(spin_bix, row, 1)

    def setup_pickups_box(self, item_pool: List[PickupEntry]):

        parent_widgets: Dict[ItemCategory, Tuple[QWidget, QGridLayout]] = {
            ItemCategory.EXPANSION: (self.expansions_box, self.expansions_layout),
            ItemCategory.ENERGY_TANK: (self.expansions_box, self.expansions_layout),
            ItemCategory.TRANSLATOR: (self.translators_box, self.translators_layout),
            ItemCategory.TEMPLE_KEY: (self.keys_box, self.keys_layout),
            ItemCategory.SKY_TEMPLE_KEY: (self.keys_box, self.keys_layout),
        }
        major_pickup_parent_widgets = (self.upgrades_box, self.upgrades_layout)

        row_for_parent = {
            self.expansions_box: 0,
            self.translators_box: 0,
            self.upgrades_box: 0,
            self.keys_box: 0,
        }
        column_for_parent = {
            self.translators_box: 0,
            self.upgrades_box: 0,
            self.keys_box: 0,
        }
        k_column_count = 2

        pickup_by_name = {}
        pickup_with_quantity = {}

        for pickup in item_pool:
            if pickup.name in pickup_by_name:
                pickup_with_quantity[pickup_by_name[pickup.name]] += 1
            else:
                pickup_by_name[pickup.name] = pickup
                pickup_with_quantity[pickup] = 1

        non_expansions_with_quantity = []

        for pickup, quantity in pickup_with_quantity.items():
            self._collected_pickups[pickup] = 0
            parent_widget, parent_layout = parent_widgets.get(pickup.item_category, major_pickup_parent_widgets)

            row = row_for_parent[parent_widget]

            if parent_widget is self.expansions_box:
                self._create_widgets_with_quantity(pickup, parent_widget, parent_layout, row, quantity)
                row_for_parent[parent_widget] += 1
            else:
                if quantity > 1:
                    non_expansions_with_quantity.append((parent_widget, parent_layout, pickup, quantity))
                else:
                    check_box = QCheckBox(parent_widget)
                    check_box.setText(pickup.name)
                    check_box.stateChanged.connect(functools.partial(self._change_item_quantity, pickup, True))
                    self._widget_for_pickup[pickup] = check_box

                    column = column_for_parent[parent_widget]
                    parent_layout.addWidget(check_box, row, column)
                    column += 1

                    if column >= k_column_count:
                        column = 0
                        row += 1

                    row_for_parent[parent_widget] = row
                    column_for_parent[parent_widget] = column

        for parent_widget, parent_layout, pickup, quantity in non_expansions_with_quantity:
            if column_for_parent[parent_widget] != 0:
                column_for_parent[parent_widget] = 0
                row_for_parent[parent_widget] += 1

            self._create_widgets_with_quantity(pickup, parent_widget, parent_layout,
                                               row_for_parent[parent_widget],
                                               quantity)
            row_for_parent[parent_widget] += 1

    def state_for_current_configuration(self) -> Optional[State]:
        all_nodes = self.game_description.world_list.all_nodes

        state = self._initial_state.copy()
        if self._actions:
            state.node = self._actions[-1]

        for teleporter, combo in self._elevator_id_to_combo.items():
            assert combo.currentData() is not None
            state.patches.elevator_connection[teleporter] = combo.currentData()

        for gate, item in self._translator_gate_to_combo.items():
            state.patches.translator_gates[gate] = item.currentData()

        for pickup, quantity in self._collected_pickups.items():
            for _ in range(quantity):
                add_pickup_to_state(state, pickup)

        for node in self._collected_nodes:
            add_resource_gain_to_current_resources(node.resource_gain_on_collect(state.patches, state.resources,
                                                                                 all_nodes),
                                                   state.resources)

        return state



from typing import Iterable

from randovania.dol_patching.assembler.ppc import Instruction, BaseInstruction


def assemble_instructions(address: int, instructions: Iterable[BaseInstruction]) -> Iterable[int]:
    for i, instruction in enumerate(instructions):
        data = list(instruction.bytes_for(address))
        yield from data
        address += len(data)


def byte_count(instructions: Iterable[BaseInstruction]) -> int:
    return sum(instruction.byte_count for instruction in instructions)



from unittest.mock import MagicMock

from PySide2.QtWidgets import QWidget

from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.games.game import RandovaniaGame
from randovania.gui.dialog import trick_details_popup
from randovania.layout.trick_level import LayoutTrickLevel


def test_click_on_link(echoes_game_description,
                       skip_qtbot):
    # Setup
    main_window = QWidget()
    main_window.open_data_visualizer_at = MagicMock()
    skip_qtbot.add_widget(main_window)
    world_name = "World"
    area_name = "Area"

    # Run
    popup = trick_details_popup.TrickDetailsPopup(
        main_window,
        main_window,
        echoes_game_description,
        TrickResourceInfo(-1, "Nothing", "Nothing", "Some description!"),
        LayoutTrickLevel.EXPERT
    )
    popup._on_click_link_to_data_editor(f"data-editor://{world_name}/{area_name}")

    # Assert
    main_window.open_data_visualizer_at.assert_called_once_with(world_name, area_name, game=RandovaniaGame.PRIME2)



import asyncio
import locale
import logging.config
import logging.handlers
import os
import sys
import traceback
import typing
from argparse import ArgumentParser
from pathlib import Path

from PySide2 import QtCore, QtWidgets
from asyncqt import asyncClose

logger = logging.getLogger(__name__)


def display_exception(val: Exception):
    if not isinstance(val, KeyboardInterrupt):
        box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Critical,
                                    "An exception was raised",
                                    "An unhandled Exception occurred:\n{}".format(val),
                                    QtWidgets.QMessageBox.Ok)
        from randovania.gui.lib import common_qt_lib
        common_qt_lib.set_default_window_icon(box)
        if val.__traceback__ is not None:
            box.setDetailedText("".join(traceback.format_tb(val.__traceback__)))
        box.exec_()


def catch_exceptions(t, val, tb):
    display_exception(val)
    old_hook(t, val, tb)


old_hook = sys.excepthook


def catch_exceptions_async(loop, context):
    if 'future' in context:
        future: asyncio.Future = context['future']
        logger.exception(context["message"], exc_info=future.exception())
    else:
        logger.critical(str(context))


async def show_main_window(app: QtWidgets.QApplication, options, is_preview: bool):
    from randovania.interface_common.preset_manager import PresetManager
    preset_manager = PresetManager(options.data_dir)

    await preset_manager.load_user_presets()

    from randovania.gui.main_window import MainWindow
    main_window = MainWindow(options, preset_manager, app.network_client, is_preview)
    app.main_window = main_window

    logger.info("Displaying main window")
    main_window.show()
    await main_window.request_new_data()


def show_tracker(app: QtWidgets.QApplication):
    from randovania.gui.auto_tracker_window import AutoTrackerWindow

    app.tracker = AutoTrackerWindow(app.game_connection, _load_options())
    logger.info("Displaying auto tracker")
    app.tracker.show()


def show_game_details(app: QtWidgets.QApplication, options, game: Path):
    from randovania.layout.layout_description import LayoutDescription
    from randovania.gui.seed_details_window import SeedDetailsWindow

    layout = LayoutDescription.from_file(game)
    details_window = SeedDetailsWindow(None, options)
    details_window.update_layout_description(layout)
    logger.info("Displaying game details")
    details_window.show()
    app.details_window = details_window


async def display_window_for(app, options, command: str, args):
    if command == "tracker":
        show_tracker(app)
    elif command == "main":
        await show_main_window(app, options, args.preview)
    elif command == "game":
        show_game_details(app, options, args.rdvgame)
    else:
        raise RuntimeError(f"Unknown command: {command}")


def create_backend(debug_game_backend: bool, options):
    from randovania.interface_common.options import Options
    options = typing.cast(Options, options)

    if debug_game_backend:
        from randovania.gui.debug_backend_window import DebugBackendWindow
        backend = DebugBackendWindow()
        backend.show()
    else:
        try:
            from randovania.game_connection.dolphin_backend import DolphinBackend
        except ImportError:
            from randovania.gui.lib import common_qt_lib
            common_qt_lib.show_install_visual_cpp_redist()
            raise SystemExit(1)

        from randovania.game_connection.nintendont_backend import NintendontBackend
        from randovania.game_connection.backend_choice import GameBackendChoice

        if options.game_backend == GameBackendChoice.NINTENDONT and options.nintendont_ip is not None:
            backend = NintendontBackend(options.nintendont_ip)
        else:
            backend = DolphinBackend()

    return backend


def _load_options():
    from randovania.interface_common.options import Options
    from randovania.gui.lib import startup_tools, theme

    options = Options.with_default_data_dir()
    if not startup_tools.load_options_from_disk(options):
        raise SystemExit(1)

    theme.set_dark_theme(options.dark_mode)

    from randovania.layout.preset_migration import VersionedPreset
    for old_preset in options.data_dir.joinpath("presets").glob("*.randovania_preset"):
        old_preset.rename(old_preset.with_name(f"{old_preset.stem}.{VersionedPreset.file_extension()}"))

    return options


def start_logger(data_dir: Path, is_preview: bool):
    # Ensure the log dir exists early on
    log_dir = data_dir.joinpath("logs")
    log_dir.mkdir(parents=True, exist_ok=True)

    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s] [%(levelname)s] [%(name)s] %(funcName)s: %(message)s',
            }
        },
        'handlers': {
            'default': {
                'level': 'DEBUG' if is_preview else 'INFO',
                'formatter': 'default',
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stdout',  # Default is stderr
            },
            'local_app_data': {
                'level': 'DEBUG',
                'formatter': 'default',
                'class': 'logging.handlers.TimedRotatingFileHandler',
                'filename': log_dir.joinpath(f"logger.log"),
                'encoding': 'utf-8',
                'backupCount': 10,
            }
        },
        'loggers': {
            'randovania.network_client.network_client': {
                'level': 'DEBUG',
            },
            'randovania.game_connection.connection_backend': {
                'level': 'DEBUG',
            },
            'randovania.gui.multiworld_client': {
                'level': 'DEBUG',
            },
            'NintendontBackend': {
                'level': 'DEBUG',
            },
            'DolphinBackend': {
                'level': 'DEBUG',
            },
            'randovania.gui.qt': {
                'level': 'INFO',
            },
            # 'socketio.client': {
            #     'level': 'DEBUG',
            # }
        },
        'root': {
            'level': 'WARNING',
            'handlers': ['default', 'local_app_data'],
        },
    })


def create_loop(app: QtWidgets.QApplication) -> asyncio.AbstractEventLoop:
    os.environ['QT_API'] = "PySide2"
    import asyncqt
    loop: asyncio.AbstractEventLoop = asyncqt.QEventLoop(app)
    asyncio.set_event_loop(loop)

    sys.excepthook = catch_exceptions
    loop.set_exception_handler(catch_exceptions_async)
    return loop


async def qt_main(app: QtWidgets.QApplication, data_dir: Path, args):
    from randovania.gui.lib.qt_network_client import QtNetworkClient
    from randovania.game_connection.game_connection import GameConnection

    app.network_client = QtNetworkClient(data_dir)
    options = _load_options()

    backend = create_backend(args.debug_game_backend, options)
    app.game_connection = GameConnection(backend)

    @asyncClose
    async def _on_last_window_closed():
        await app.network_client.disconnect_from_server()
        await app.game_connection.stop()
        logger.info("Last QT window closed")

    app.lastWindowClosed.connect(_on_last_window_closed, QtCore.Qt.QueuedConnection)

    await asyncio.gather(app.game_connection.start(),
                         display_window_for(app, options, args.command, args))


def run(args):
    locale.setlocale(locale.LC_ALL, "")  # use system's default locale
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

    data_dir = args.custom_network_storage
    if data_dir is None:
        from randovania.interface_common import persistence
        data_dir = persistence.user_data_dir()

    is_preview = args.preview
    start_logger(data_dir, is_preview)
    app = QtWidgets.QApplication(sys.argv)

    def main_done(done: asyncio.Task):
        e: typing.Optional[Exception] = done.exception()
        if e is not None:
            display_exception(e)

    loop = create_loop(app)
    with loop:
        loop.create_task(qt_main(app, data_dir, args)).add_done_callback(main_done)
        loop.run_forever()


def create_subparsers(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "gui",
        help="Run the Graphical User Interface"
    )
    parser.add_argument("--preview", action="store_true", help="Activates preview features")
    parser.add_argument("--custom-network-storage", type=Path, help="Use a custom path to store the network login.")
    parser.add_argument("--debug-game-backend", action="store_true", help="Opens the debug game backend.")

    gui_parsers = parser.add_subparsers(dest="command")
    gui_parsers.add_parser("main", help="Displays the Main Window").set_defaults(func=run)
    gui_parsers.add_parser("tracker", help="Opens only the auto tracker").set_defaults(func=run)

    game_parser = gui_parsers.add_parser("game", help="Opens an rdvgame")
    game_parser.add_argument("rdvgame", type=Path, help="Path ")
    game_parser.set_defaults(func=run)

    def check_command(args):
        if args.command is None:
            parser.print_help()
            raise SystemExit(1)

    parser.set_defaults(func=check_command)



import copy
import dataclasses
from dataclasses import dataclass
from typing import Dict, Tuple, Iterator

from randovania.game_description.area_location import AreaLocation
from randovania.game_description.assignment import PickupAssignment, GateAssignment, PickupTarget
from randovania.game_description.dock import DockWeakness, DockConnection
from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.game_description.hint import Hint
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import CurrentResources
from randovania.game_description.resources.resource_type import ResourceType


@dataclass(frozen=True)
class GamePatches:
    """Determines patches that are made to the game's data.
    Currently we support:
    * Swapping pickup locations
    """
    player_index: int
    pickup_assignment: PickupAssignment
    elevator_connection: Dict[int, AreaLocation]
    dock_connection: Dict[Tuple[int, int], DockConnection]
    dock_weakness: Dict[Tuple[int, int], DockWeakness]
    translator_gates: GateAssignment
    starting_items: CurrentResources
    starting_location: AreaLocation
    hints: Dict[LogbookAsset, Hint]
    game_specific: EchoesGameSpecific

    def assign_new_pickups(self, assignments: Iterator[Tuple[PickupIndex, PickupTarget]]) -> "GamePatches":
        new_pickup_assignment = copy.copy(self.pickup_assignment)

        for index, pickup in assignments:
            assert index not in new_pickup_assignment
            new_pickup_assignment[index] = pickup

        return dataclasses.replace(self, pickup_assignment=new_pickup_assignment)

    def assign_pickup_assignment(self, assignment: PickupAssignment) -> "GamePatches":
        items: Iterator[Tuple[PickupIndex, PickupTarget]] = assignment.items()
        return self.assign_new_pickups(items)

    def assign_gate_assignment(self, assignment: GateAssignment) -> "GamePatches":
        new_translator_gates = copy.copy(self.translator_gates)

        for gate, translator in assignment.items():
            assert gate not in new_translator_gates
            assert gate.resource_type == ResourceType.GATE_INDEX
            new_translator_gates[gate] = translator

        return dataclasses.replace(self, translator_gates=new_translator_gates)

    def assign_starting_location(self, location: AreaLocation) -> "GamePatches":
        return dataclasses.replace(self, starting_location=location)

    def assign_extra_initial_items(self, new_resources: CurrentResources) -> "GamePatches":
        current = copy.copy(self.starting_items)

        for resource, quantity in new_resources.items():
            if resource.resource_type != ResourceType.ITEM:
                raise ValueError("Only ITEM is supported as extra initial items, got {}".format(resource.resource_type))
            current[resource] = current.get(resource, 0) + quantity

        return dataclasses.replace(self, starting_items=current)

    def assign_hint(self, logbook: LogbookAsset, hint: Hint) -> "GamePatches":
        current = copy.copy(self.hints)
        current[logbook] = hint
        return dataclasses.replace(self, hints=current)



from dataclasses import dataclass
from enum import unique, Enum
from typing import NamedTuple, List

from randovania.game_description.requirements import Requirement


@unique
class DockType(Enum):
    DOOR = 0
    MORPH_BALL_DOOR = 1
    OTHER = 2
    PORTAL = 3


@dataclass(frozen=True, order=True)
class DockWeakness:
    index: int
    name: str
    is_blast_shield: bool
    requirement: Requirement
    dock_type: DockType

    def __hash__(self):
        return hash((self.index, self.name, self.dock_type))

    def __repr__(self):
        return self.name

    @property
    def long_name(self):
        return self.name


def _find_dock_weakness_with_id(info_list: List[DockWeakness],
                                index: int) -> DockWeakness:
    for info in info_list:
        if info.index == index:
            return info
    raise ValueError(
        "Dock weakness with index {} not found in {}".format(index, info_list))


class DockWeaknessDatabase(NamedTuple):
    door: List[DockWeakness]
    morph_ball: List[DockWeakness]
    other: List[DockWeakness]
    portal: List[DockWeakness]

    def get_by_type(self, dock_type: DockType) -> List[DockWeakness]:
        if dock_type == DockType.DOOR:
            return self.door
        elif dock_type == DockType.MORPH_BALL_DOOR:
            return self.morph_ball
        elif dock_type == DockType.OTHER:
            return self.other
        elif dock_type == DockType.PORTAL:
            return self.portal
        else:
            raise ValueError("Invalid dock_type: {}".format(dock_type))

    def get_by_type_and_index(self, dock_type: DockType,
                              weakness_index: int) -> DockWeakness:
        return _find_dock_weakness_with_id(
            self.get_by_type(dock_type), weakness_index)


@dataclass(frozen=True, order=True)
class DockConnection:
    area_asset_id: int
    dock_index: int

    def __repr__(self):
        return "{}/{}".format(self.area_asset_id, self.dock_index)



"""
Some constants related to specific Echoes items that the logic needs to deal with.
"""

DARK_TEMPLE_KEY_MODEL = 37
DARK_TEMPLE_KEY_NAMES = [
    "Dark Agon Key {0}",
    "Dark Torvus Key {0}",
    "Ing Hive Key {0}"
]
DARK_TEMPLE_KEY_ITEMS = [
    [32, 33, 34, ],
    [35, 36, 37, ],
    [38, 39, 40, ],
]
SKY_TEMPLE_KEY_MODEL = 38
SKY_TEMPLE_KEY_ITEMS = [
    29,
    30,
    31,
    101,
    102,
    103,
    104,
    105,
    106,
]
USELESS_PICKUP_MODEL = 30
USELESS_PICKUP_ITEM = 107



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'item_configuration_popup.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_ItemConfigurationPopup(object):
    def setupUi(self, ItemConfigurationPopup):
        if not ItemConfigurationPopup.objectName():
            ItemConfigurationPopup.setObjectName(u"ItemConfigurationPopup")
        ItemConfigurationPopup.resize(503, 63)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ItemConfigurationPopup.sizePolicy().hasHeightForWidth())
        ItemConfigurationPopup.setSizePolicy(sizePolicy)
        ItemConfigurationPopup.setMaximumSize(QSize(16777215, 16777215))
        self.gridLayout_2 = QGridLayout(ItemConfigurationPopup)
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.item_name_label = QLabel(ItemConfigurationPopup)
        self.item_name_label.setObjectName(u"item_name_label")

        self.gridLayout_2.addWidget(self.item_name_label, 0, 0, 1, 1)

        self.vanilla_radio = QRadioButton(ItemConfigurationPopup)
        self.vanilla_radio.setObjectName(u"vanilla_radio")

        self.gridLayout_2.addWidget(self.vanilla_radio, 0, 2, 1, 1)

        self.starting_radio = QRadioButton(ItemConfigurationPopup)
        self.starting_radio.setObjectName(u"starting_radio")

        self.gridLayout_2.addWidget(self.starting_radio, 0, 3, 1, 1)

        self.shuffled_spinbox = QSpinBox(ItemConfigurationPopup)
        self.shuffled_spinbox.setObjectName(u"shuffled_spinbox")
        self.shuffled_spinbox.setMinimum(1)
        self.shuffled_spinbox.setMaximum(99)

        self.gridLayout_2.addWidget(self.shuffled_spinbox, 0, 5, 1, 1)

        self.shuffled_radio = QRadioButton(ItemConfigurationPopup)
        self.shuffled_radio.setObjectName(u"shuffled_radio")

        self.gridLayout_2.addWidget(self.shuffled_radio, 0, 4, 1, 1)

        self.excluded_radio = QRadioButton(ItemConfigurationPopup)
        self.excluded_radio.setObjectName(u"excluded_radio")

        self.gridLayout_2.addWidget(self.excluded_radio, 0, 1, 1, 1)

        self.provided_ammo_spinbox = QSpinBox(ItemConfigurationPopup)
        self.provided_ammo_spinbox.setObjectName(u"provided_ammo_spinbox")

        self.gridLayout_2.addWidget(self.provided_ammo_spinbox, 1, 4, 1, 2)

        self.provided_ammo_label = QLabel(ItemConfigurationPopup)
        self.provided_ammo_label.setObjectName(u"provided_ammo_label")
        self.provided_ammo_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)
        self.provided_ammo_label.setWordWrap(True)

        self.gridLayout_2.addWidget(self.provided_ammo_label, 1, 0, 1, 4)


        self.retranslateUi(ItemConfigurationPopup)

        QMetaObject.connectSlotsByName(ItemConfigurationPopup)
    # setupUi

    def retranslateUi(self, ItemConfigurationPopup):
        ItemConfigurationPopup.setWindowTitle(QCoreApplication.translate("ItemConfigurationPopup", u"Item Configuration", None))
        self.item_name_label.setText(QCoreApplication.translate("ItemConfigurationPopup", u"TextLabel", None))
        self.vanilla_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Vanilla", None))
        self.starting_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Starting", None))
        self.shuffled_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Shuffled", None))
        self.excluded_radio.setText(QCoreApplication.translate("ItemConfigurationPopup", u"Excluded", None))
#if QT_CONFIG(tooltip)
        self.provided_ammo_label.setToolTip(QCoreApplication.translate("ItemConfigurationPopup", u"<html><head/><body><p>When this item is collected, it also gives this amount of the given ammos.</p><p>This is included in the calculation of how much each pickup of this ammo gives.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.provided_ammo_label.setText(QCoreApplication.translate("ItemConfigurationPopup", u"<html><head/><body><p>Provided Ammo (XXXX and YYYY)</p></body></html>", None))
    # retranslateUi




import struct

import pytest

from randovania.dol_patching.dol_file import DolFile


@pytest.fixture()
def dol_file(tmp_path):
    section_size = 450
    data = bytearray(b"\x00" * (0x100 + section_size))
    data[0:4] = struct.pack(">L", 0x100)
    data[0x48:0x48 + 4] = struct.pack(">L", 0x2000)
    data[0x90:0x90 + 4] = struct.pack(">L", section_size)

    tmp_path.joinpath("test.dol").write_bytes(data)
    dol_file = DolFile(tmp_path.joinpath("test.dol"))
    return dol_file



import datetime
import logging
import random
from functools import partial
from pathlib import Path
from typing import Optional

from PySide2.QtWidgets import QDialog, QMessageBox, QWidget, QMenu, QAction
from asyncqt import asyncSlot

from randovania.games.game import RandovaniaGame
from randovania.gui.generated.main_window_ui import Ui_MainWindow
from randovania.gui.lib import preset_describer, common_qt_lib, async_dialog
from randovania.gui.lib.background_task_mixin import BackgroundTaskMixin
from randovania.gui.lib.generation_failure_handling import GenerationFailureHandler
from randovania.gui.lib.window_manager import WindowManager
from randovania.gui.preset_settings.logic_settings_window import LogicSettingsWindow
from randovania.interface_common import simplified_patcher
from randovania.interface_common.options import Options
from randovania.interface_common.preset_editor import PresetEditor
from randovania.interface_common.status_update_lib import ProgressUpdateCallable
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink
from randovania.layout.preset import Preset
from randovania.layout.preset_migration import VersionedPreset, InvalidPreset
from randovania.resolver.exceptions import GenerationFailure


def persist_layout(data_dir: Path, description: LayoutDescription):
    history_dir = data_dir.joinpath("game_history")
    history_dir.mkdir(parents=True, exist_ok=True)

    date_format = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    file_path = history_dir.joinpath(
        f"{date_format}-{description.shareable_word_hash}.{description.file_extension()}")
    description.save_to_file(file_path)


class GenerateSeedTab(QWidget, BackgroundTaskMixin):
    _logic_settings_window: Optional[LogicSettingsWindow] = None
    _has_preset: bool = False
    _tool_button_menu: QMenu
    _action_delete: QAction

    def __init__(self, window: Ui_MainWindow, window_manager: WindowManager, options: Options):
        super().__init__()

        self.window = window
        self._window_manager = window_manager
        self.failure_handler = GenerationFailureHandler(self)
        self._options = options

    def setup_ui(self):
        window = self.window

        # Progress
        self.background_tasks_button_lock_signal.connect(self.enable_buttons_with_background_tasks)
        self.progress_update_signal.connect(self.update_progress)
        self.window.stop_background_process_button.clicked.connect(self.stop_background_process)

        for game in RandovaniaGame:
            self.window.create_choose_game_combo.addItem(game.long_name, game)

        self.window.create_choose_game_combo.setVisible(self._window_manager.is_preview_mode)
        self.window.create_choose_game_label.setVisible(self._window_manager.is_preview_mode)
        self.window.num_players_spin_box.setVisible(self._window_manager.is_preview_mode)

        # Menu
        self._tool_button_menu = QMenu(window.preset_tool_button)
        window.preset_tool_button.setMenu(self._tool_button_menu)

        self._action_delete = QAction(window)
        self._action_delete.setText("Delete")
        self._tool_button_menu.addAction(self._action_delete)

        action_export_preset = QAction(window)
        action_export_preset.setText("Export")
        self._tool_button_menu.addAction(action_export_preset)

        action_import_preset = QAction(window)
        action_import_preset.setText("Import")
        self._tool_button_menu.addAction(action_import_preset)

        # Signals
        window.create_choose_game_combo.activated.connect(self._on_select_game)
        window.preset_tool_button.clicked.connect(self._on_customize_button)
        window.create_preset_combo.activated.connect(self._on_select_preset)
        window.create_generate_button.clicked.connect(partial(self._generate_new_seed, True))
        window.create_generate_race_button.clicked.connect(partial(self._generate_new_seed, False))

        self._action_delete.triggered.connect(self._on_delete_preset)
        action_export_preset.triggered.connect(self._on_export_preset)
        action_import_preset.triggered.connect(self._on_import_preset)

    @property
    def _current_preset_data(self) -> Optional[VersionedPreset]:
        return self._window_manager.preset_manager.preset_for_name(self.window.create_preset_combo.currentData())

    def enable_buttons_with_background_tasks(self, value: bool):
        self.window.stop_background_process_button.setEnabled(not value)
        self.window.create_generate_button.setEnabled(value)
        self.window.create_generate_race_button.setEnabled(value)

    def _create_button_for_preset(self, preset: VersionedPreset):
        create_preset_combo = self.window.create_preset_combo
        create_preset_combo.addItem(preset.name, preset.name)

    def _add_new_preset(self, preset: VersionedPreset):
        with self._options as options:
            options.selected_preset_name = preset.name

        if self._window_manager.preset_manager.add_new_preset(preset):
            self._create_button_for_preset(preset)
        self.on_preset_changed(preset.get_preset())

    @asyncSlot()
    async def _on_customize_button(self):
        if self._logic_settings_window is not None:
            self._logic_settings_window.raise_()
            return

        editor = PresetEditor(self._current_preset_data.get_preset())
        self._logic_settings_window = LogicSettingsWindow(self._window_manager, editor)

        self._logic_settings_window.on_preset_changed(editor.create_custom_preset_with())
        editor.on_changed = lambda: self._logic_settings_window.on_preset_changed(editor.create_custom_preset_with())

        result = await async_dialog.execute_dialog(self._logic_settings_window)
        self._logic_settings_window = None

        if result == QDialog.Accepted:
            self._add_new_preset(VersionedPreset.with_preset(editor.create_custom_preset_with()))

    def _on_delete_preset(self):
        self._window_manager.preset_manager.delete_preset(self._current_preset_data)
        self.window.create_preset_combo.removeItem(self.window.create_preset_combo.currentIndex())
        self._on_select_preset()

    def _on_export_preset(self):
        path = common_qt_lib.prompt_user_for_preset_file(self._window_manager, new_file=True)
        if path is not None:
            self._current_preset_data.save_to_file(path)

    def _on_import_preset(self):
        path = common_qt_lib.prompt_user_for_preset_file(self._window_manager, new_file=False)
        if path is None:
            return

        preset = VersionedPreset.from_file_sync(path)
        try:
            preset.get_preset()
        except (ValueError, KeyError):
            QMessageBox.critical(
                self._window_manager,
                "Error loading preset",
                "The file at '{}' contains an invalid preset.".format(path)
            )
            return

        if self._window_manager.preset_manager.preset_for_name(preset.name) is not None:
            user_response = QMessageBox.warning(
                self._window_manager,
                "Preset name conflict",
                "A preset named '{}' already exists. Do you want to overwrite it?".format(preset.name),
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No
            )
            if user_response == QMessageBox.No:
                return

        self._add_new_preset(preset)

    def select_game(self, game: RandovaniaGame):
        combo_index = self.window.create_choose_game_combo.findData(game)
        self.window.create_choose_game_combo.setCurrentIndex(combo_index)
        self._update_create_preset_combo(game)

    def _on_select_game(self):
        game = self.window.create_choose_game_combo.currentData()
        self._update_create_preset_combo(game)
        self._on_select_preset()

    def _update_create_preset_combo(self, game: RandovaniaGame):
        self.window.create_preset_combo.clear()
        for preset in self._window_manager.preset_manager.all_presets:
            if preset.game == game:
                self._create_button_for_preset(preset)

    def _on_select_preset(self):
        preset_data = self._current_preset_data
        try:
            self.on_preset_changed(preset_data.get_preset())
        except InvalidPreset as e:
            logging.exception(f"Invalid preset for {preset_data.name}")
            QMessageBox.warning(
                self._window_manager,
                "Incompatible Preset",
                f"Preset {preset_data.name} can't be used as it contains the following error:\n{e.original_exception}"
            )
            self.window.create_preset_combo.setCurrentIndex(0)
            self.on_preset_changed(self._window_manager.preset_manager.default_preset.get_preset())
            return

        with self._options as options:
            options.selected_preset_name = preset_data.name

    # Generate seed

    def _generate_new_seed(self, spoiler: bool):
        preset = self._current_preset_data
        num_players = self.window.num_players_spin_box.value()

        self.generate_seed_from_permalink(Permalink(
            seed_number=random.randint(0, 2 ** 31),
            spoiler=spoiler,
            presets={
                i: preset.get_preset()
                for i in range(num_players)
            },
        ))

    def generate_seed_from_permalink(self, permalink: Permalink):
        def work(progress_update: ProgressUpdateCallable):
            try:
                layout = simplified_patcher.generate_layout(progress_update=progress_update,
                                                            permalink=permalink,
                                                            options=self._options)
                progress_update(f"Success! (Seed hash: {layout.shareable_hash})", 1)
                persist_layout(self._options.data_dir, layout)
                self._window_manager.open_game_details(layout)

            except GenerationFailure as generate_exception:
                self.failure_handler.handle_failure(generate_exception)
                progress_update("Generation Failure: {}".format(generate_exception), -1)

        if self._window_manager.is_preview_mode:
            print(f"Permalink: {permalink.as_base64_str}")
        self.run_in_background_thread(work, "Creating a seed...")

    def on_options_changed(self, options: Options):
        if not self._has_preset:
            selected_preset = self._window_manager.preset_manager.preset_for_name(options.selected_preset_name)
            if selected_preset is not None:
                self.select_game(selected_preset.game)
                index = self.window.create_preset_combo.findText(selected_preset.name)
                if index != -1:
                    self.window.create_preset_combo.setCurrentIndex(index)
                    try:
                        self.on_preset_changed(self._current_preset_data.get_preset())
                        return
                    except InvalidPreset:
                        logging.exception(f"Invalid preset for {options.selected_preset_name}")
            else:
                self.select_game(RandovaniaGame.PRIME2)

            self.window.create_preset_combo.setCurrentIndex(0)
            self.on_preset_changed(self._window_manager.preset_manager.default_preset.get_preset())

    def on_preset_changed(self, preset: Preset):
        self._has_preset = True

        self.window.create_preset_description.setText(preset.description)
        self._action_delete.setEnabled(preset.base_preset_name is not None)

        create_preset_combo = self.window.create_preset_combo
        create_preset_combo.setCurrentIndex(create_preset_combo.findText(preset.name))

        categories = list(preset_describer.describe(preset))
        left_categories = categories[::2]
        right_categories = categories[1::2]

        self.window.create_describe_left_label.setText(preset_describer.merge_categories(left_categories))
        self.window.create_describe_right_label.setText(preset_describer.merge_categories(right_categories))

    def update_progress(self, message: str, percentage: int):
        self.window.progress_label.setText(message)
        if "Aborted" in message:
            percentage = 0
        if percentage >= 0:
            self.window.progress_bar.setRange(0, 100)
            self.window.progress_bar.setValue(percentage)
        else:
            self.window.progress_bar.setRange(0, 0)



import datetime

import pytest
from PySide2.QtWidgets import QDialog
from mock import patch, AsyncMock, MagicMock

from randovania.gui.online_game_list_window import GameSessionBrowserDialog
from randovania.network_client.game_session import GameSessionListEntry
from randovania.network_common.session_state import GameSessionState


@pytest.mark.asyncio
@patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
async def test_attempt_join(mock_execute_dialog: AsyncMock,
                            skip_qtbot):
    # Setup
    utc = datetime.timezone.utc
    mock_execute_dialog.return_value = QDialog.Accepted
    network_client = MagicMock()
    network_client.join_game_session = AsyncMock()
    session_a = GameSessionListEntry(id=1, name="A Game", has_password=True, state=GameSessionState.FINISHED,
                                     num_players=1, creator="You",
                                     creation_date=datetime.datetime(year=2015, month=5, day=1, tzinfo=utc))
    session_b = GameSessionListEntry(id=2, name="B Game", has_password=True, state=GameSessionState.IN_PROGRESS,
                                     num_players=1, creator="You",
                                     creation_date=datetime.datetime.now(utc) - datetime.timedelta(days=4))

    dialog = GameSessionBrowserDialog(network_client)
    dialog.sessions = [session_a, session_b]
    dialog.update_list()
    dialog.table_widget.selectRow(0)

    # Run
    await dialog.attempt_join()

    # Assert
    mock_execute_dialog.assert_awaited_once()
    network_client.join_game_session.assert_awaited_once_with(session_b, "")



from typing import Callable, List

ProgressUpdateCallable = Callable[[str, float], None]


class ConstantPercentageCallback:
    def __init__(self, status_update: ProgressUpdateCallable, percentage: float):
        self.status_update = status_update
        self.percentage = percentage

    def __call__(self, message: str) -> None:
        self.status_update(message, self.percentage)


class SuccessiveCallback:
    def __init__(self, status_update: ProgressUpdateCallable, max_calls: int):
        self.call_count = 0
        self.status_update = status_update
        self.max_calls = max_calls

    def __call__(self, message: str) -> None:
        self.call_count += 1
        self.status_update(
            message,
            min(self.call_count, self.max_calls) / self.max_calls
        )


class OffsetProgressUpdate:
    def __init__(self, status_update: ProgressUpdateCallable, offset: float, scale: float):
        self.status_update = status_update
        self.offset = offset
        self.scale = scale

    def __call__(self, message: str, percentage: float) -> None:
        new_value = percentage if percentage < 0 else self.offset + percentage * self.scale
        self.status_update(
            message,
            new_value
        )


def create_progress_update_from_successive_messages(
        status_update: ProgressUpdateCallable,
        max_calls: int) -> Callable[[str], None]:
    """
    Creates a callable that invokes the given ProgressUpdateCallable each it's called,
    with percentage based on call count.
    :param status_update:
    :param max_calls:
    :return:
    """
    return SuccessiveCallback(status_update, max_calls)


def split_progress_update(status_update: ProgressUpdateCallable, parts: int) -> List[ProgressUpdateCallable]:
    """
    Creates a list of ProgressUpdateCallable that when called progress successively,
    progresses a given ProgressUpdateCallable.
    :param status_update:
    :param parts:
    :return:
    """
    scale = 1.0 / parts
    return [
        OffsetProgressUpdate(status_update, scale * i, scale)
        for i in range(parts)
    ]



from unittest.mock import MagicMock, patch

import pytest

from randovania.cli import prime_database
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo


@pytest.mark.parametrize("expected_resource", [0, 1, 2])
@patch("randovania.cli.prime_database._list_paths_with_resource", autospec=True)
@patch("randovania.cli.prime_database.load_game_description", autospec=True)
def test_list_paths_with_resource_logic(mock_load_game_description: MagicMock,
                                        mock_list_paths_with_resource: MagicMock,
                                        expected_resource: int
                                        ):
    # Setup
    args = MagicMock()
    game = mock_load_game_description.return_value

    resource_a = SimpleResourceInfo(0, "Long Name A", "A", ResourceType.ITEM)
    resource_b = SimpleResourceInfo(0, "Long Name B", "B", ResourceType.ITEM)
    resource_c = SimpleResourceInfo(0, "Long Name C", "C", ResourceType.ITEM)
    resources = [resource_a, resource_b, resource_c]

    game.resource_database = [
        [resource_a],
        [resource_b],
        [resource_c],
    ]

    resource = resources[expected_resource]
    args.resource = resource.long_name

    # Run
    prime_database.list_paths_with_resource_logic(args)

    # Assert
    mock_load_game_description.assert_called_once_with(args)
    mock_list_paths_with_resource.assert_called_once_with(
        game,
        args.print_only_area,
        resource,
        None
    )



from random import Random
from typing import Dict, List, Iterator

import randovania
from randovania.game_description import data_reader
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.assignment import GateAssignment, PickupTarget
from randovania.game_description.default_database import default_prime2_memo_data
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.node import TeleporterNode, PickupNode
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceGainTuple, ResourceGain, CurrentResources, \
    ResourceInfo
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.world_list import WorldList
from randovania.games.prime.patcher_file_lib import sky_temple_key_hint, item_hints
from randovania.generator.item_pool import pickup_creator, pool_creator
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.hint_configuration import HintConfiguration, SkyTempleKeyHintMode
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.elevators import LayoutElevators
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.pickup_model import PickupModelStyle, PickupModelDataSource

_EASTER_EGG_RUN_VALIDATED_CHANCE = 1024
_EASTER_EGG_SHINY_MISSILE = 8192
_CUSTOM_NAMES_FOR_ELEVATORS = {
    # Great Temple
    408633584: "Temple Transport Emerald",
    2399252740: "Temple Transport Violet",
    2556480432: "Temple Transport Amber",

    # Temple Grounds to Great Temple
    1345979968: "Sanctuary Quadrant",
    1287880522: "Agon Quadrant",
    2918020398: "Torvus Quadrant",

    # Temple Grounds to Areas
    1660916974: "Agon Gate",
    2889020216: "Torvus Gate",
    3455543403: "Sanctuary Gate",

    # Agon
    1473133138: "Agon Entrance",
    2806956034: "Agon Portal Access",
    3331021649: "Agon Temple Access",

    # Torvus
    1868895730: "Torvus Entrance",
    3479543630: "Torvus Temple Access",
    3205424168: "Lower Torvus Access",

    # Sanctuary
    3528156989: "Sanctuary Entrance",
    900285955: "Sanctuary Spider side",
    3145160350: "Sanctuary Vault side",
}

_RESOURCE_NAME_TRANSLATION = {
    'Temporary Missile': 'Missile',
    'Temporary Power Bombs': 'Power Bomb',
}


def _resource_user_friendly_name(resource: ResourceInfo) -> str:
    """
    Gets a name that we should display to the user for the given resource.
    :param resource:
    :return:
    """
    return _RESOURCE_NAME_TRANSLATION.get(resource.long_name, resource.long_name)


def _create_spawn_point_field(patches: GamePatches,
                              resource_database: ResourceDatabase,
                              ) -> dict:
    capacities = [
        {
            "index": item.index,
            "amount": patches.starting_items.get(item, 0),
        }
        for item in resource_database.item
    ]

    return {
        "location": patches.starting_location.as_json,
        "amount": capacities,
        "capacity": capacities,
    }


def _get_jingle_index_for(category: ItemCategory) -> int:
    if category.is_key:
        return 2
    elif category.is_major_category and category != ItemCategory.ENERGY_TANK:
        return 1
    else:
        return 0


def _pickup_scan(pickup: PickupEntry) -> str:
    if pickup.item_category != ItemCategory.EXPANSION:
        if len(pickup.resources) > 1 and all(conditional.name is not None for conditional in pickup.resources):
            return "{}. Provides the following in order: {}".format(
                pickup.name, ", ".join(conditional.name for conditional in pickup.resources))
        else:
            return pickup.name

    return "{} that provides {}".format(
        pickup.name,
        " and ".join(
            "{} {}".format(quantity, _resource_user_friendly_name(resource))
            for resource, quantity in pickup.resources[-1].resources
        )
    )


def _create_pickup_resources_for(resources: ResourceGain):
    return [
        {
            "index": resource.index,
            "amount": quantity
        }
        for resource, quantity in resources
        if quantity > 0 and resource.resource_type == ResourceType.ITEM
    ]


def _get_single_hud_text(pickup_name: str,
                         memo_data: Dict[str, str],
                         resources: ResourceGainTuple,
                         ) -> str:
    return memo_data[pickup_name].format(**{
        _resource_user_friendly_name(resource): quantity
        for resource, quantity in resources
    })


def _get_all_hud_text(pickup: PickupEntry,
                      memo_data: Dict[str, str],
                      ) -> List[str]:
    return [
        _get_single_hud_text(conditional.name or pickup.name, memo_data, conditional.resources)
        for conditional in pickup.resources
    ]


def _calculate_hud_text(pickup: PickupEntry,
                        visual_pickup: PickupEntry,
                        model_style: PickupModelStyle,
                        memo_data: Dict[str, str],
                        ) -> List[str]:
    """
    Calculates what the hud_text for a pickup should be
    :param pickup:
    :param visual_pickup:
    :param model_style:
    :param memo_data:
    :return:
    """

    if model_style == PickupModelStyle.HIDE_ALL:
        hud_text = _get_all_hud_text(visual_pickup, memo_data)
        if len(hud_text) == len(pickup.resources):
            return hud_text
        else:
            return [hud_text[0]] * len(pickup.resources)

    else:
        return _get_all_hud_text(pickup, memo_data)


class PickupCreator:
    def __init__(self, rng: Random):
        self.rng = rng

    def create_pickup_data(self,
                           original_index: PickupIndex,
                           pickup_target: PickupTarget,
                           visual_pickup: PickupEntry,
                           model_style: PickupModelStyle,
                           scan_text: str) -> dict:
        raise NotImplementedError()

    def create_pickup(self,
                      original_index: PickupIndex,
                      pickup_target: PickupTarget,
                      visual_pickup: PickupEntry,
                      model_style: PickupModelStyle,
                      ) -> dict:
        model_pickup = pickup_target.pickup if model_style == PickupModelStyle.ALL_VISIBLE else visual_pickup

        if model_style in {PickupModelStyle.ALL_VISIBLE, PickupModelStyle.HIDE_MODEL}:
            scan_text = _pickup_scan(pickup_target.pickup)
        else:
            scan_text = visual_pickup.name

        # TODO: less improvised, really
        model_index = model_pickup.model_index
        if model_index == 22 and self.rng.randint(0, _EASTER_EGG_SHINY_MISSILE) == 0:
            # If placing a missile expansion model, replace with Dark Missile Trooper model with a 1/8192 chance
            model_index = 23

        result = {
            "pickup_index": original_index.index,
            **self.create_pickup_data(original_index, pickup_target, visual_pickup, model_style, scan_text),
            "model_index": model_index,
            "sound_index": 1 if model_pickup.item_category.is_key else 0,
            "jingle_index": _get_jingle_index_for(model_pickup.item_category),
        }
        return result


class PickupCreatorSolo(PickupCreator):
    def __init__(self, rng: Random, memo_data: Dict[str, str]):
        super().__init__(rng)
        self.memo_data = memo_data

    def create_pickup_data(self,
                           original_index: PickupIndex,
                           pickup_target: PickupTarget,
                           visual_pickup: PickupEntry,
                           model_style: PickupModelStyle,
                           scan_text: str) -> dict:
        hud_text = _calculate_hud_text(pickup_target.pickup, visual_pickup, model_style, self.memo_data)
        if hud_text == ["Energy Transfer Module acquired!"] and (
                self.rng.randint(0, _EASTER_EGG_RUN_VALIDATED_CHANCE) == 0):
            hud_text = ["Run validated!"]

        return {
            "resources": _create_pickup_resources_for(pickup_target.pickup.resources[0].resources),
            "conditional_resources": [
                {
                    "item": conditional.item.index,
                    "resources": _create_pickup_resources_for(conditional.resources),
                }
                for conditional in pickup_target.pickup.resources[1:]
            ],
            "convert": [
                {
                    "from_item": conversion.source.index,
                    "to_item": conversion.target.index,
                    "clear_source": conversion.clear_source,
                    "overwrite_target": conversion.overwrite_target,
                }
                for conversion in pickup_target.pickup.convert_resources
            ],
            "hud_text": hud_text,
            "scan": scan_text,
        }


class PickupCreatorMulti(PickupCreator):
    def __init__(self, rng: Random, memo_data: Dict[str, str], players_config: PlayersConfiguration):
        super().__init__(rng)
        self.solo_creator = PickupCreatorSolo(rng, memo_data)
        self.players_config = players_config

    def create_pickup_data(self,
                           original_index: PickupIndex,
                           pickup_target: PickupTarget,
                           visual_pickup: PickupEntry,
                           model_style: PickupModelStyle,
                           scan_text: str) -> dict:
        if pickup_target.player == self.players_config.player_index:
            result = self.solo_creator.create_pickup_data(original_index, pickup_target, visual_pickup,
                                                          model_style, scan_text)
            result["scan"] = f"Your {result['scan']}"
        else:
            other_name = self.players_config.player_names[pickup_target.player]
            result: dict = {
                "resources": [],
                "conditional_resources": [],
                "convert": [],
                "hud_text": [f"Sent {pickup_target.pickup.name} to {other_name}!"],
                "scan": f"{other_name}'s {scan_text}",
            }

        magic_resource = {
            "index": 74,
            "amount": original_index.index + 1,
        }

        result["resources"].append(magic_resource)
        for conditional in result["conditional_resources"]:
            conditional["resources"].append(magic_resource)

        return result


def _get_visual_model(original_index: int,
                      pickup_list: List[PickupTarget],
                      data_source: PickupModelDataSource,
                      ) -> PickupEntry:
    if data_source == PickupModelDataSource.ETM:
        return pickup_creator.create_visual_etm()
    elif data_source == PickupModelDataSource.RANDOM:
        return pickup_list[original_index % len(pickup_list)].pickup
    elif data_source == PickupModelDataSource.LOCATION:
        raise NotImplementedError()
    else:
        raise ValueError(f"Unknown data_source: {data_source}")


def _create_pickup_list(patches: GamePatches,
                        useless_target: PickupTarget,
                        pickup_count: int,
                        rng: Random,
                        model_style: PickupModelStyle,
                        data_source: PickupModelDataSource,
                        creator: PickupCreator,
                        ) -> list:
    """
    Creates the patcher data for all pickups in the game
    :param patches:
    :param useless_target:
    :param pickup_count:
    :param rng:
    :param model_style:
    :param data_source:
    :param creator:
    :return:
    """
    pickup_assignment = patches.pickup_assignment

    pickup_list = list(pickup_assignment.values())
    rng.shuffle(pickup_list)

    pickups = [
        creator.create_pickup(PickupIndex(i),
                              pickup_assignment.get(PickupIndex(i), useless_target),
                              _get_visual_model(i, pickup_list, data_source),
                              model_style,
                              )
        for i in range(pickup_count)
    ]

    return pickups


def _elevator_area_name(world_list: WorldList,
                        area_location: AreaLocation,
                        include_world_name: bool,
                        ) -> str:
    if area_location.area_asset_id in _CUSTOM_NAMES_FOR_ELEVATORS:
        return _CUSTOM_NAMES_FOR_ELEVATORS[area_location.area_asset_id]

    else:
        world = world_list.world_by_area_location(area_location)
        area = world.area_by_asset_id(area_location.area_asset_id)
        if include_world_name:
            return world_list.area_name(area)
        else:
            return area.name


def _pretty_name_for_elevator(world_list: WorldList,
                              original_teleporter_node: TeleporterNode,
                              connection: AreaLocation,
                              ) -> str:
    """
    Calculates the name the room that contains this elevator should have
    :param world_list:
    :param original_teleporter_node:
    :param connection:
    :return:
    """
    if original_teleporter_node.keep_name_when_vanilla:
        if original_teleporter_node.default_connection == connection:
            return world_list.nodes_to_area(original_teleporter_node).name

    return "Transport to {}".format(_elevator_area_name(world_list, connection, False))


def _create_elevators_field(patches: GamePatches, game: GameDescription) -> list:
    """
    Creates the elevator entries in the patcher file
    :param patches:
    :param game:
    :return:
    """

    world_list = game.world_list
    nodes_by_teleporter_id = _get_nodes_by_teleporter_id(world_list)
    elevator_connection = patches.elevator_connection

    if len(elevator_connection) != len(nodes_by_teleporter_id):
        raise ValueError("Invalid elevator count. Expected {}, got {}.".format(
            len(nodes_by_teleporter_id), len(elevator_connection)
        ))

    elevators = [
        {
            "instance_id": instance_id,
            "origin_location": world_list.node_to_area_location(nodes_by_teleporter_id[instance_id]).as_json,
            "target_location": connection.as_json,
            "room_name": _pretty_name_for_elevator(world_list, nodes_by_teleporter_id[instance_id], connection)
        }
        for instance_id, connection in elevator_connection.items()
    ]

    return elevators


def _get_nodes_by_teleporter_id(world_list: WorldList) -> Dict[int, TeleporterNode]:
    nodes_by_teleporter_id = {
        node.teleporter_instance_id: node

        for node in world_list.all_nodes
        if isinstance(node, TeleporterNode) and node.editable
    }
    return nodes_by_teleporter_id


def _create_translator_gates_field(gate_assignment: GateAssignment) -> list:
    """
    Creates the translator gate entries in the patcher file
    :param gate_assignment:
    :return:
    """
    return [
        {
            "gate_index": gate.index,
            "translator_index": translator.index,
        }
        for gate, translator in gate_assignment.items()
    ]


def _apply_translator_gate_patches(specific_patches: dict, elevators: LayoutElevators) -> None:
    """

    :param specific_patches:
    :param elevators:
    :return:
    """
    specific_patches["always_up_gfmc_compound"] = True
    specific_patches["always_up_torvus_temple"] = True
    specific_patches["always_up_great_temple"] = elevators != LayoutElevators.VANILLA


def _create_elevator_scan_port_patches(world_list: WorldList, elevator_connection: Dict[int, AreaLocation],
                                       ) -> Iterator[dict]:
    nodes_by_teleporter_id = _get_nodes_by_teleporter_id(world_list)

    for teleporter_id, node in nodes_by_teleporter_id.items():
        if node.scan_asset_id is None:
            continue

        target_area_name = _elevator_area_name(world_list, elevator_connection[teleporter_id], True)
        yield {
            "asset_id": node.scan_asset_id,
            "strings": [f"Access to &push;&main-color=#FF3333;{target_area_name}&pop; granted.", ""],
        }


def _logbook_title_string_patches():
    return [
        {
            "asset_id": 3271034066,
            "strings": [
                'Hints', 'Violet', 'Cobalt', 'Technology', 'Keys 1, 2, 3', 'Keys 7, 8, 9', 'Regular Hints',
                'Emerald', 'Amber', '&line-spacing=75;Flying Ing\nCache Hints', 'Keys 4, 5, 6', 'Keys 1, 2, 3',
                '&line-spacing=75;Torvus Energy\nController', 'Underground Tunnel', 'Training Chamber',
                'Catacombs', 'Gathering Hall', '&line-spacing=75;Fortress\nTransport\nAccess',
                '&line-spacing=75;Hall of Combat\nMastery', 'Main Gyro Chamber',
                '&line-spacing=75;Sanctuary\nEnergy\nController', 'Main Research', 'Watch Station',
                'Sanctuary Entrance', '&line-spacing=75;Transport to\nAgon Wastes', 'Mining Plaza',
                '&line-spacing=75;Agon Energy\nController', 'Portal Terminal', 'Mining Station B',
                'Mining Station A', 'Meeting Grounds', 'Path of Eyes', 'Path of Roots',
                '&line-spacing=75;Main Energy\nController', "A-Kul's Testament",
                '&line-spacing=75;Central\nMining\nStation', 'Main Reactor', 'Torvus Lagoon', 'Catacombs',
                'Sanctuary Entrance', "Dynamo Works", 'Storage Cavern A', 'Landing Site', 'Industrial Site',
                '&line-spacing=75;Sky Temple\nKey Hints', 'Keys 7, 8, 9', 'Keys 4, 5, 6', 'Sky Temple Key 1',
                'Sky Temple Key 2', 'Sky Temple Key 3', 'Sky Temple Key 4', 'Sky Temple Key 5',
                'Sky Temple Key 6', 'Sky Temple Key 7', 'Sky Temple Key 8', 'Sky Temple Key 9'
            ],
        }, {
            "asset_id": 2301408881,
            "strings": [
                'Research', 'Mechanisms', 'Luminoth Technology', 'Biology', 'GF Security', 'Vehicles',
                'Aether Studies', 'Aether', 'Dark Aether', 'Phazon', 'Sandgrass', 'Blueroot Tree',
                'Ing Webtrap',
                'Webling', 'U-Mos', 'Bladepod', 'Ing Storage', 'Flying Ing Cache', 'Torvus Bearerpod',
                'Agon Bearerpod', 'Ingworm Cache', 'Ingsphere Cache', 'Plantforms', 'Darklings',
                'GF Gate Mk VI',
                'GF Gate Mk VII', 'GF Lock Mk V', 'GF Defense Shield', 'Kinetic Orb Cannon', 'GF Bridge',
                "Samus's Gunship", 'GFS Tyr', 'Pirate Skiff', 'Visors', 'Weapon Systems', 'Armor',
                'Morph Ball Systems', 'Movement Systems', 'Beam Weapons', 'Scan Visor', 'Combat Visor',
                'Dark Visor',
                'Echo Visor', 'Morph Ball', 'Boost Ball', 'Spider Ball', 'Morph Ball Bomb', 'Power Bomb',
                'Dark Bomb', 'Light Bomb', 'Annihilator Bomb', 'Space Jump Boots', 'Screw Attack',
                'Gravity Boost',
                'Grapple Beam', 'Varia Suit', 'Dark Suit', 'Light Suit', 'Power Beam', 'Dark Beam',
                'Light Beam',
                'Annihilator Beam', 'Missile Launcher', 'Seeker Missile Launcher', 'Super Missile',
                'Sonic Boom',
                'Darkburst', 'Sunburst', 'Charge Beam', 'Missile Systems', 'Charge Combos', 'Morph Balls',
                'Bomb Systems', 'Miscellaneous', 'Dark Temple Keys', 'Bloatsac', 'Luminoth Technology',
                'Light Beacons', 'Light Crystals', 'Lift Crystals', 'Utility Crystals', 'Light Crystal',
                'Energized Crystal', 'Nullified Crystal', 'Super Crystal', 'Light Beacon', 'Energized Beacon',
                'Nullified Beacon', 'Super Beacon', 'Inactive Beacon', 'Dark Lift Crystal',
                'Light Lift Crystal',
                'Liftvine Crystal', 'Torvus Hanging Pod', 'Sentinel Crystal', 'Dark Sentinel Crystal',
                'Systems',
                'Bomb Slot', 'Spinner', 'Grapple Point', 'Spider Ball Track', 'Energy Tank',
                'Beam Ammo Expansion',
                'Missile Expansion', 'Dark Agon Keys', 'Dark Torvus Keys', 'Ing Hive Keys', 'Sky Temple Keys',
                'Temple Grounds', 'Sanctuary Fortress', 'Torvus Bog', 'Agon Wastes', 'Dark Agon Temple Key 1',
                'Dark Agon Temple Key 2', 'Dark Agon Temple Key 3', 'Dark Torvus Temple Key 1',
                'Dark Torvus Temple Key 2', 'Dark Torvus Temple Key 3', 'Ing Hive Temple Key 1',
                'Ing Hive Temple Key 2', 'Ing Hive Temple Key 3', 'Sky Temple Key 1', 'Sky Temple Key 2',
                'Sky Temple Key 3', 'Sky Temple Key 4', 'Sky Temple Key 5', 'Sky Temple Key 6',
                'Sky Temple Key 7',
                'Sky Temple Key 8', 'Sky Temple Key 9', 'Suit Expansions', 'Charge Combo', 'Ingclaw',
                'Dormant Ingclaw', 'Power Bomb Expansion', 'Energy Transfer Module', 'Cocoons',
                'Splinter Cocoon',
                'War Wasp Hive', 'Metroid Cocoon', 'Dark Aether', 'Aether', 'Dark Portal', 'Light Portal',
                'Energy Controller', 'Wall Jump Surface',
            ]
        },
    ]


def _create_string_patches(hint_config: HintConfiguration,
                           game: GameDescription,
                           all_patches: Dict[int, GamePatches],
                           players_config: PlayersConfiguration,
                           rng: Random,
                           ) -> list:
    """

    :param hint_config:
    :param game:
    :param patches:
    :return:
    """
    patches = all_patches[players_config.player_index]
    string_patches = []

    # Location Hints
    string_patches.extend(
        item_hints.create_hints(all_patches, players_config, game.world_list, rng)
    )

    # Sky Temple Keys
    stk_mode = hint_config.sky_temple_keys
    if stk_mode == SkyTempleKeyHintMode.DISABLED:
        string_patches.extend(sky_temple_key_hint.hide_hints())
    else:
        string_patches.extend(sky_temple_key_hint.create_hints(all_patches, players_config, game.world_list,
                                                               stk_mode == SkyTempleKeyHintMode.HIDE_AREA))

    # Elevator Scans
    string_patches.extend(_create_elevator_scan_port_patches(game.world_list, patches.elevator_connection))

    string_patches.extend(_logbook_title_string_patches())

    return string_patches


def additional_starting_items(layout_configuration: EchoesConfiguration,
                              resource_database: ResourceDatabase,
                              starting_items: CurrentResources) -> List[str]:
    initial_items = pool_creator.calculate_pool_results(layout_configuration, resource_database)[2]

    return [
        "{}{}".format("{} ".format(quantity) if quantity > 1 else "", _resource_user_friendly_name(item))
        for item, quantity in starting_items.items()
        if 0 < quantity != initial_items.get(item, 0)
    ]


def _create_starting_popup(layout_configuration: EchoesConfiguration,
                           resource_database: ResourceDatabase,
                           starting_items: CurrentResources) -> list:
    extra_items = additional_starting_items(layout_configuration, resource_database, starting_items)
    if extra_items:
        return [
            "Extra starting items:",
            ", ".join(extra_items)
        ]
    else:
        return []


class _SimplifiedMemo(dict):
    def __missing__(self, key):
        return "{} acquired!".format(key)


def _simplified_memo_data() -> Dict[str, str]:
    result = _SimplifiedMemo()
    result["Temporary Power Bombs"] = "Power Bomb Expansion acquired, but the main Power Bomb is required to use it."
    result["Temporary Missile"] = "Missile Expansion acquired, but the Missile Launcher is required to use it."
    return result


def create_patcher_file(description: LayoutDescription,
                        players_config: PlayersConfiguration,
                        cosmetic_patches: CosmeticPatches,
                        ) -> dict:
    """

    :param description:
    :param players_config:
    :param cosmetic_patches:
    :return:
    """
    preset = description.permalink.get_preset(players_config.player_index)
    configuration = preset.configuration
    patches = description.all_patches[players_config.player_index]
    rng = Random(description.permalink.seed_number)

    game = data_reader.decode_data(configuration.game_data)
    pickup_count = game.world_list.num_pickup_nodes
    useless_target = PickupTarget(pickup_creator.create_useless_pickup(game.resource_database),
                                  players_config.player_index)

    result = {}
    _add_header_data_to_result(description, result)

    result["menu_mod"] = configuration.menu_mod
    result["user_preferences"] = cosmetic_patches.user_preferences.as_json
    result["default_items"] = {
        "visor": configuration.major_items_configuration.default_items[ItemCategory.VISOR].name,
        "beam": configuration.major_items_configuration.default_items[ItemCategory.BEAM].name,
    }

    # Add Spawn Point
    result["spawn_point"] = _create_spawn_point_field(patches, game.resource_database)

    result["starting_popup"] = _create_starting_popup(configuration, game.resource_database, patches.starting_items)

    # Add the pickups
    if cosmetic_patches.disable_hud_popup:
        memo_data = _simplified_memo_data()
    else:
        memo_data = default_prime2_memo_data()

    if description.permalink.player_count == 1:
        creator = PickupCreatorSolo(rng, memo_data)
    else:
        creator = PickupCreatorMulti(rng, memo_data, players_config)

    result["pickups"] = _create_pickup_list(patches,
                                            useless_target, pickup_count,
                                            rng,
                                            configuration.pickup_model_style,
                                            configuration.pickup_model_data_source,
                                            creator=creator,
                                            )

    # Add the elevators
    result["elevators"] = _create_elevators_field(patches, game)

    # Add translators
    result["translator_gates"] = _create_translator_gates_field(patches.translator_gates)

    # Scan hints
    result["string_patches"] = _create_string_patches(configuration.hints, game, description.all_patches, players_config, rng)

    # TODO: if we're starting at ship, needs to collect 9 sky temple keys and want item loss,
    # we should disable hive_chamber_b_post_state
    result["specific_patches"] = {
        "hive_chamber_b_post_state": True,
        "intro_in_post_state": True,
        "warp_to_start": configuration.warp_to_start,
        "speed_up_credits": cosmetic_patches.speed_up_credits,
        "disable_hud_popup": cosmetic_patches.disable_hud_popup,
        "pickup_map_icons": cosmetic_patches.pickup_markers,
        "full_map_at_start": cosmetic_patches.open_map,
        "dark_world_varia_suit_damage": configuration.varia_suit_damage,
        "dark_world_dark_suit_damage": configuration.dark_suit_damage,
    }

    result["logbook_patches"] = [
        {"asset_id": 25, "connections": [81, 166, 195], },
        {"asset_id": 38, "connections": [4, 33, 120, 251, 364], },
        {"asset_id": 60, "connections": [38, 74, 154, 196], },
        {"asset_id": 74, "connections": [59, 75, 82, 102, 260], },
        {"asset_id": 81, "connections": [148, 151, 156], },
        {"asset_id": 119, "connections": [60, 254, 326], },
        {"asset_id": 124, "connections": [35, 152, 355], },
        {"asset_id": 129, "connections": [29, 118, 367], },
        {"asset_id": 154, "connections": [169, 200, 228, 243, 312, 342], },
        {"asset_id": 166, "connections": [45, 303, 317], },
        {"asset_id": 194, "connections": [1, 6], },
        {"asset_id": 195, "connections": [159, 221, 231], },
        {"asset_id": 196, "connections": [17, 19, 23, 162, 183, 379], },
        {"asset_id": 233, "connections": [58, 191, 373], },
        {"asset_id": 241, "connections": [223, 284], },
        {"asset_id": 254, "connections": [129, 233, 319], },
        {"asset_id": 318, "connections": [119, 216, 277, 343], },
        {"asset_id": 319, "connections": [52, 289, 329], },
        {"asset_id": 326, "connections": [124, 194, 241, 327], },
        {"asset_id": 327, "connections": [46, 275], },
    ]

    _apply_translator_gate_patches(result["specific_patches"], configuration.elevators)

    return result


def _add_header_data_to_result(description: LayoutDescription, result: dict) -> None:
    result["permalink"] = "-permalink-"
    result["seed_hash"] = f"- {description.shareable_word_hash} ({description.shareable_hash})"
    result["randovania_version"] = randovania.VERSION



import functools
import json
from pathlib import Path

from randovania import get_data_path
from randovania.game_description import data_reader
from randovania.game_description.data_reader import read_resource_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.item import item_database
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.games.game import RandovaniaGame
from randovania.games.prime import default_data


def default_prime2_game_description() -> GameDescription:
    return game_description_for(RandovaniaGame.PRIME2)


@functools.lru_cache()
def resource_database_for(game: RandovaniaGame) -> ResourceDatabase:
    return read_resource_database(default_data.read_json_then_binary(game)[1]["resource_database"])


def game_description_for(game: RandovaniaGame) -> GameDescription:
    return data_reader.decode_data(default_data.read_json_then_binary(game)[1])


def _read_database_in_path(path: Path) -> item_database.ItemDatabase:
    configuration_path = path.joinpath("configuration")

    with configuration_path.joinpath("major-items.json").open() as major_items_file:
        major_items_data = json.load(major_items_file)

    with configuration_path.joinpath("ammo.json").open() as ammo_file:
        ammo_data = json.load(ammo_file)

    return item_database.read_database(major_items_data, ammo_data)


@functools.lru_cache()
def default_prime1_item_database() -> item_database.ItemDatabase:
    return _read_database_in_path(get_data_path().joinpath("item_database", "prime1"))


@functools.lru_cache()
def default_prime2_item_database() -> item_database.ItemDatabase:
    return _read_database_in_path(get_data_path().joinpath("item_database", "prime2"))


@functools.lru_cache()
def default_prime3_item_database() -> item_database.ItemDatabase:
    return _read_database_in_path(get_data_path().joinpath("item_database", "prime3"))


def item_database_for_game(game: RandovaniaGame):
    if game == RandovaniaGame.PRIME1:
        return default_prime1_item_database()
    elif game == RandovaniaGame.PRIME2:
        return default_prime2_item_database()
    elif game == RandovaniaGame.PRIME3:
        return default_prime3_item_database()
    else:
        raise ValueError(f"Unknown game: {game}")


@functools.lru_cache()
def default_prime2_memo_data() -> dict:
    with get_data_path().joinpath("item_database", "prime2", "memo_data.json").open("r") as memo_data_file:
        memo_data = json.load(memo_data_file)

    item_database.add_memo_data_keys(memo_data)

    return memo_data



import dataclasses
from typing import Iterable

from randovania.dol_patching.dol_file import DolFile
from randovania.games.game import RandovaniaGame


@dataclasses.dataclass(frozen=True)
class DolVersion:
    game: RandovaniaGame
    description: str
    build_string_address: int
    build_string: bytes
    sda2_base: int


def find_version_for_dol(dol_file: DolFile, all_versions: Iterable[DolVersion]) -> DolVersion:
    dol_file.set_editable(False)
    with dol_file:
        for version in all_versions:
            build_string = dol_file.read(version.build_string_address, len(version.build_string))
            if build_string == version.build_string:
                return version

    raise RuntimeError(f"Unsupported game version")



import argparse
import struct


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("hexstring")
    args = parser.parse_args()

    b = bytes.fromhex(args.hexstring)

    op_type, num_ops, num_addresses, keep_alive = struct.unpack_from(">BBBB", b, 0)
    print(f"> Operation type: {op_type}")
    print(f"> Keep Alive: {bool(keep_alive)}")
    print(f"> Num addresses: {num_addresses}")

    if op_type == 1:
        return

    offset = 4
    addresses = struct.unpack_from(f">{num_addresses}I", b, offset)

    for i, address in enumerate(addresses):
        print(f"  * Address {i}: 0x{address:08x}")

    offset += num_addresses * 4

    print(f"> Num operations: {num_ops}")
    for i in range(num_ops):
        op_byte = struct.unpack_from(">B", b, offset)[0]
        offset += 1

        operation_pretty = []

        if bool(op_byte & 0x20):
            byte_count = 4
        else:
            byte_count = struct.unpack_from(">B", b, offset)[0]
            offset += 1

        # Has read
        if bool(op_byte & 0x80):
            operation_pretty.append(f"read {byte_count} bytes")

        address = addresses[op_byte & 0x7]
        address_text = f"0x{address:08x}"

        # Has offset
        if bool(op_byte & 0x10):
            address_offset = struct.unpack_from(">h", b, offset)[0]
            offset += 2
            address_text = f"*{address_text} {address_offset:+05x}"

        # Has write
        if bool(op_byte & 0x40):
            write_data = b[offset:offset + byte_count]
            offset += byte_count
            if len(write_data) != byte_count:
                raise ValueError(f"At op {i}, expected {byte_count} bytes. Got {len(write_data)}")
            operation_pretty.append(f"write {write_data.hex()}")
            if byte_count == 4:
                operation_pretty[-1] += f" ({int.from_bytes(write_data, 'big')})"

        print(f"  * At {address_text}, {' and '.join(operation_pretty)}.")


if __name__ == '__main__':
    main()



import contextlib

import pytest
from flask_socketio import ConnectionRefusedError

from randovania.server import app


def test_create_app(mocker, tmpdir):
    mocker.patch("randovania.get_configuration").return_value = {
        "discord_client_id": 1234,
        "guest_key": "s2D-pjBIXqEqkbeRvkapeDn82MgZXLLQGZLTgqqZ--A=",
        "server_config": {
            "secret_key": "key",
            "discord_client_secret": 5678,
            "fernet_key": 's2D-pjBIXqEqkbeRvkapeDn82MgZXLLQGZLTgqqZ--A=',
            "database_path": str(tmpdir.join("database.db")),
            "client_version_checking": "strict",
        }
    }
    mock_game_session = mocker.patch("randovania.server.game_session.setup_app")
    mock_user_session = mocker.patch("randovania.server.user_session.setup_app")
    mock_create_sio = mocker.patch("flask_socketio.SocketIO")

    # Run
    result = app.create_app()

    # Assert
    mock_game_session.assert_called_once_with(result.sio)
    mock_user_session.assert_called_once_with(result.sio)
    mock_create_sio.assert_called_once_with(result)
    assert tmpdir.join("database.db").exists()

    with result.test_client() as test_client:
        assert test_client.get("/").data.decode("utf-8") == "ok"

    assert result.config['SECRET_KEY'] == "key"
    assert result.config["DISCORD_CLIENT_ID"] == 1234
    assert result.config["DISCORD_CLIENT_SECRET"] == 5678
    assert result.config["DISCORD_REDIRECT_URI"] == "http://127.0.0.1:5000/callback/"
    assert result.config["FERNET_KEY"] == b's2D-pjBIXqEqkbeRvkapeDn82MgZXLLQGZLTgqqZ--A='

    encrpyted_value = b'gAAAAABfSh6fY4FOiqfGWMHXdE9A4uNVEu5wfn8BAsgP8EZ0-f-lqbYDqYzdiblhT5xhk-wMmG8sOLgKNN-dUaiV7n6JCydn7Q=='
    assert result.sio.fernet_encrypt.decrypt(encrpyted_value) == b'banana'


@pytest.mark.parametrize(["mode", "client_version", "server_version", "expected"], [
    (app.ClientVersionCheck.STRICT, "1.0", "1.0", True),
    (app.ClientVersionCheck.STRICT, "1.0", "1.0.1", False),
    (app.ClientVersionCheck.STRICT, "1.0", "1.1", False),
    (app.ClientVersionCheck.MATCH_MAJOR_MINOR, "1.0", "1.0", True),
    (app.ClientVersionCheck.MATCH_MAJOR_MINOR, "1.0", "1.0.1", True),
    (app.ClientVersionCheck.MATCH_MAJOR_MINOR, "1.0", "1.1", False),
    (app.ClientVersionCheck.IGNORE, "1.0", "1.0", True),
    (app.ClientVersionCheck.IGNORE, "1.0", "1.0.1", True),
    (app.ClientVersionCheck.IGNORE, "1.0", "1.1", True),
])
def test_check_client_version(mode, client_version, server_version, expected):
    if expected:
        expectation = contextlib.nullcontext()
    else:
        expectation = pytest.raises(ConnectionRefusedError)

    with expectation:
        app.check_client_version(mode, client_version, server_version)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'connections_editor.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_ConnectionEditor(object):
    def setupUi(self, ConnectionEditor):
        if not ConnectionEditor.objectName():
            ConnectionEditor.setObjectName(u"ConnectionEditor")
        ConnectionEditor.resize(794, 393)
        ConnectionEditor.setMinimumSize(QSize(600, 0))
        self.gridLayout_2 = QGridLayout(ConnectionEditor)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.visualizer_scroll = QScrollArea(ConnectionEditor)
        self.visualizer_scroll.setObjectName(u"visualizer_scroll")
        self.visualizer_scroll.setWidgetResizable(True)
        self.visualizer_contents = QWidget()
        self.visualizer_contents.setObjectName(u"visualizer_contents")
        self.visualizer_contents.setGeometry(QRect(0, 0, 774, 345))
        self.contents_layout = QVBoxLayout(self.visualizer_contents)
        self.contents_layout.setObjectName(u"contents_layout")
        self.visualizer_scroll.setWidget(self.visualizer_contents)

        self.gridLayout_2.addWidget(self.visualizer_scroll, 0, 0, 1, 1)

        self.button_box = QDialogButtonBox(ConnectionEditor)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setOrientation(Qt.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.gridLayout_2.addWidget(self.button_box, 1, 0, 1, 1)


        self.retranslateUi(ConnectionEditor)
        self.button_box.accepted.connect(ConnectionEditor.accept)
        self.button_box.rejected.connect(ConnectionEditor.reject)

        QMetaObject.connectSlotsByName(ConnectionEditor)
    # setupUi

    def retranslateUi(self, ConnectionEditor):
        ConnectionEditor.setWindowTitle(QCoreApplication.translate("ConnectionEditor", u"Connection Editor", None))
    # retranslateUi




from typing import Dict, Iterator, List

from randovania.game_description.item.ammo import Ammo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.generator.item_pool.pickup_creator import create_ammo_expansion
from randovania.layout.ammo_configuration import AmmoConfiguration
from randovania.layout.ammo_state import AmmoState
from randovania.resolver.exceptions import InvalidConfiguration


def add_ammo(resource_database: ResourceDatabase,
             ammo_configuration: AmmoConfiguration,
             included_ammo_for_item: Dict[int, int],
             ) -> Iterator[PickupEntry]:
    """
    Creates the necessary pickups for the given ammo_configuration.
    :param resource_database:
    :param ammo_configuration:
    :param included_ammo_for_item: How much of each item was provided based on the major items.
    :return:
    """

    previous_pickup_for_item = {}

    for ammo, state in ammo_configuration.items_state.items():
        if state.pickup_count == 0:
            continue

        if state.variance != 0:
            raise InvalidConfiguration("Variance was configured for {0.name}, but it is currently NYI".format(ammo))

        ammo_per_pickup = items_for_ammo(ammo, state,
                                         included_ammo_for_item,
                                         previous_pickup_for_item,
                                         ammo_configuration.maximum_ammo)

        for ammo_quantity in ammo_per_pickup:
            yield create_ammo_expansion(ammo, ammo_quantity, state.requires_major_item, resource_database)


def items_for_ammo(ammo: Ammo,
                   state: AmmoState,
                   included_ammo_for_item: Dict[int, int],
                   previous_pickup_for_item: Dict[int, Ammo],
                   maximum_ammo: Dict[int, int],
                   ) -> List[List[int]]:
    """
    Helper function for add_ammo.

    :param maximum_ammo:
    :param ammo:
    :param state:
    :param included_ammo_for_item:
    :param previous_pickup_for_item:
    :return: An array that lists how many of each ammo each instance of the expansions should give
    """
    ammo_per_pickup: List[List[int]] = [[] for _ in range(state.pickup_count)]

    # TODO: add test for case of 0 pickups
    if state.pickup_count == 0:
        return ammo_per_pickup

    for item in ammo.items:
        if item in previous_pickup_for_item:
            raise InvalidConfiguration(
                "Both {0.name} and {1.name} have non-zero pickup count for item {2}. This is unsupported.".format(
                    ammo, previous_pickup_for_item[item], item)
            )
        previous_pickup_for_item[item] = ammo

        if item not in included_ammo_for_item:
            raise InvalidConfiguration(
                "Invalid configuration: ammo {0.name} was configured for {1.pickup_count}"
                "expansions, but main pickup was removed".format(ammo, state)
            )

        if maximum_ammo[item] < included_ammo_for_item[item]:
            raise InvalidConfiguration(
                "Ammo {0.name} was configured for a total of {1}, but major items gave {2}".format(
                    ammo, maximum_ammo[item], included_ammo_for_item[item]))

        adjusted_count = maximum_ammo[item] - included_ammo_for_item[item]
        count_per_expansion = adjusted_count // state.pickup_count
        expansions_with_extra_count = adjusted_count - count_per_expansion * state.pickup_count

        for i, pickup_ammo in enumerate(ammo_per_pickup):
            pickup_ammo.append(count_per_expansion)
            if i < expansions_with_extra_count:
                pickup_ammo[-1] += 1

    return ammo_per_pickup



import time
from typing import Set

from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import Node, PickupNode
from randovania.game_description.requirements import RequirementList, RequirementSet
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.generator.generator_reach import GeneratorReach, get_collectable_resource_nodes_of_reach
from randovania.resolver.logic import Logic

_DEBUG_LEVEL = 0
count = 0
_current_indent = 0
_last_printed_additional: dict = None


def n(node: Node, world_list, with_world=False) -> str:
    return world_list.node_name(node, with_world) if node is not None else "None"


def sorted_requirementset_print(new_requirements: Set[RequirementList]):
    to_print = []
    for requirement in new_requirements:
        to_print.append(", ".join(str(item) for item in sorted(requirement.values())))
    print("\n".join(x for x in sorted(to_print)))


def increment_attempts():
    global count
    count += 1
    # if count > 1500:
    #     raise SystemExit


def _indent(offset=0):
    return " " * (_current_indent - offset)


def log_resolve_start():
    global _current_indent, _last_printed_additional
    _current_indent = 0
    _last_printed_additional = {}


def log_new_advance(state: "State", reach: "ResolverReach"):
    from randovania.resolver.state import State
    from randovania.resolver.resolver_reach import ResolverReach
    state: State
    reach: ResolverReach

    global _current_indent
    increment_attempts()
    _current_indent += 1
    if _DEBUG_LEVEL > 0:
        world_list = state.world_list

        if hasattr(state.node, "resource"):
            resource = state.node.resource()
            if isinstance(resource, PickupIndex):
                resource = state.patches.pickup_assignment.get(resource)
                if resource is not None:
                    resource = resource.pickup
        else:
            resource = None

        print("{}> {} for {}".format(_indent(1), n(state.node, world_list=world_list), resource))
        if _DEBUG_LEVEL >= 3:
            for node in reach.nodes:
                print("{}: {}".format(_indent(), n(node, world_list=world_list)))


def log_checking_satisfiable_actions():
    if _DEBUG_LEVEL > 1:
        print("{}# Satisfiable Actions".format(_indent()))


def log_rollback(state: "State", has_action, possible_action: bool):
    global _current_indent
    if _DEBUG_LEVEL > 1:
        print("{}* Rollback {}; Had action? {}; Possible Action? {}".format(
            _indent(),
            n(state.node, world_list=state.world_list),
            has_action, possible_action))
    _current_indent -= 1


def log_skip_action_missing_requirement(node: Node, game: GameDescription, requirement_set: RequirementSet):
    if _DEBUG_LEVEL > 1:
        if node in _last_printed_additional and _last_printed_additional[node] == requirement_set:
            print("{}* Skip {}, same additional".format(_indent(), n(node, world_list=game.world_list)))
        else:
            print("{}* Skip {}, missing additional:".format(_indent(), n(node, world_list=game.world_list)))
            requirement_set.pretty_print(_indent(-1))
            _last_printed_additional[node] = requirement_set


def print_distribute_one_item_detail(potential_pickup_nodes, start_time):
    if _DEBUG_LEVEL > 0:
        print(":: {:2d} pickups spots :: Took {}s".format(
            len(potential_pickup_nodes), time.perf_counter() - start_time
        ))


def print_distribute_one_item(state: "State", available_item_pickups):
    if _DEBUG_LEVEL > 0:
        print("\n> Distribute starting at {} with {} resources and {} pickups left.".format(
            n(state.node, world_list=state.world_list),
            len(state.resources),
            len(available_item_pickups)
        ))
        return time.perf_counter()


def print_actions_of_reach(reach: GeneratorReach):
    if _DEBUG_LEVEL <= 1:
        return

    game = reach.game
    actions = get_collectable_resource_nodes_of_reach(reach)

    for action in actions:
        print("++ Safe? {1} -- {0} -- Dangerous? {2}".format(
            game.world_list.node_name(action),
            reach.is_safe_node(action),
            action.resource() in game.dangerous_resources
        ))


def set_level(level: int):
    global _DEBUG_LEVEL
    if isinstance(level, int):
        _DEBUG_LEVEL = level
    else:
        _DEBUG_LEVEL = 0


def debug_level() -> int:
    return _DEBUG_LEVEL


def debug_print(message: str):
    if _DEBUG_LEVEL > 0:
        print(message)



import pytest

from randovania.game_description.node import GenericNode, DockNode, PickupNode, TeleporterNode, EventNode, \
    TranslatorGateNode, LogbookNode, PlayerShipNode
from randovania.gui.dialog.node_details_popup import NodeDetailsPopup


@pytest.mark.parametrize("node_type", [
    GenericNode,
    DockNode,
    PickupNode,
    TeleporterNode,
    EventNode,
    TranslatorGateNode,
    LogbookNode,
])
def test_unchanged_create_new_node_echoes(skip_qtbot, echoes_game_description, node_type):
    node = next(node for node in echoes_game_description.world_list.all_nodes if isinstance(node, node_type))
    dialog = NodeDetailsPopup(echoes_game_description, node)

    # Run
    new_node = dialog.create_new_node()

    # Assert
    assert node == new_node


@pytest.mark.parametrize("node_type", [
    PlayerShipNode,
])
def test_unchanged_create_new_node_corruption(skip_qtbot, corruption_game_description, node_type):
    node = next(node for node in corruption_game_description.world_list.all_nodes if isinstance(node, node_type))
    dialog = NodeDetailsPopup(corruption_game_description, node)

    # Run
    new_node = dialog.create_new_node()

    # Assert
    assert node == new_node



from enum import unique, Enum


@unique
class ResourceType(Enum):
    ITEM = 0
    EVENT = 1
    TRICK = 2
    DAMAGE = 3
    VERSION = 4
    MISC = 5
    PICKUP_INDEX = 7
    GATE_INDEX = 8
    LOGBOOK_INDEX = 9
    SHIP_NODE = 10

    @property
    def negated_prefix(self) -> str:
        if self is ResourceType.EVENT:
            return "Before "
        elif self is ResourceType.MISC:
            return "Disabled "
        else:
            return "No "

    @property
    def non_negated_prefix(self) -> str:
        if self is ResourceType.EVENT:
            return "After "
        elif self is ResourceType.MISC:
            return "Enabled "
        else:
            return ""

    def __lt__(self, other):
        return self.value < other.value
